/////////////////////////////////////////////////////////////////////
//
//            X   X           X
//           XXX XX         XX
//          XXXXXXXX      XXX
//         XX X XXXXXXXXXXX
//        XXXXX XXXXXXXXX
//       XXXXX XXXXXXXXXX
//            XXX XXX XXX
//           XXX XX   XX
//           X   X     X
//
//    Copyright (C) 2003-2020  Ron Jakl
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
/////////////////////////////////////////////////////////////////////

#include <QGraphicsScene>
#include <QGraphicsPixmapItem>
#include <QGraphicsItem>
#include <QMouseEvent>
#include <QPixmap>
#include <QPainter>
#include <QBrush>
#include <QPen>
#include <QColor>
#include <QPoint>
#include <QRect>
#include <QFont>
#include <QRectF>

#include "datatypes.h"
#include "displaywidget.h"

static const int						ITEM_ID_VALUE			= 0x00000001;
static const int						ITEM_ID_VALUE_FIELD		= 0x00000002;
static const int						ITEM_ID_TEXT_FIELD		= 0x00000004;
static const int						ITEM_POS_FIELD			= 0x00000008;
static const int						ITEM_BORDER_TYPE_FIELD	= 0x00000010;
static const int						ITEM_SELECTED_FIELD		= 0x00000011;

static const int						PIXMAP_BORDER_SIZE		= 5;


TDisplayWidget::TDisplayWidget(
	const QWidget						*parent)
:QGraphicsView(const_cast<QWidget*>(parent))
{
	d_graphics_scene = new QGraphicsScene(this);
	
	this->setScene(d_graphics_scene);
	this->setDragMode(QGraphicsView::NoDrag);
	
//	d_graphics_scene->setBackgroundBrush(QBrush(QColor(240,240,240),Qt::DiagCrossPattern));
	
	d_label_item_size = QSize(32,32);
	d_font_size = 12;
	
	d_background_image = d_graphics_scene->addPixmap(QPixmap(":/images/background_default.png"));
	d_background_image->setTransformationMode(Qt::SmoothTransformation);
	
	d_background_image->setZValue(0);	// base layer
	
	d_moving_item = 0;
	
	d_label_color = QColor(TDefaultColor::COLOR_LABEL_RED,TDefaultColor::COLOR_LABEL_GREEN,TDefaultColor::COLOR_LABEL_BLUE);;
	
	// create item pixmap at default size with RGBA
	QImage								item(d_label_item_size,QImage::Format_ARGB32);
	
	item.fill(Qt::transparent);
	d_item_pixmap = QPixmap::fromImage(item);
}

TDisplayWidget::~TDisplayWidget(void)
{

}

bool TDisplayWidget::Item_Exists(
	const int							id) const
{
	int									item_id;
	std::vector<QGraphicsPixmapItem*>::const_iterator item_iter;
	
	for(item_iter = d_label_item_pixmaps.begin();item_iter != d_label_item_pixmaps.end();++item_iter)
	{
		item_id = (*item_iter)->data(ITEM_ID_VALUE).toInt();
		
		if(id == item_id)
		{
			return true;
		}
	}
	
	return false;
}

QPointF TDisplayWidget::Item_Position(
	const int							id) const
{
	int									item_id;
	QPointF								center_position;
	std::vector<QGraphicsPixmapItem*>::const_iterator item_iter;
	
	for(item_iter = d_label_item_pixmaps.begin();item_iter != d_label_item_pixmaps.end();++item_iter)
	{
		item_id = (*item_iter)->data(ITEM_ID_VALUE).toInt();
		
		if(id == item_id)
		{
			center_position = (*item_iter)->data(ITEM_POS_FIELD).toPointF();
			break;
		}
	}
	
	return center_position;
}

QPixmap TDisplayWidget::Pixmap(void)
{
	QRectF								scene_rect;
	QRectF								target_rect;
	
	this->Update_Scene_Rect();

	scene_rect = d_graphics_scene->sceneRect();
	target_rect = scene_rect;
	
	target_rect.setRight(target_rect.right() - target_rect.left());
	target_rect.setBottom(target_rect.bottom() - target_rect.top());
	target_rect.setLeft(0);
	target_rect.setTop(0);

	QPixmap	pixmap(static_cast<int>(target_rect.width()),static_cast<int>(target_rect.height()));
	pixmap.fill();
	QPainter							painter(&pixmap);
	
	d_graphics_scene->render(&painter,target_rect,scene_rect);
	
	return pixmap;
}

QPixmap TDisplayWidget::Background_Pixmap(void) const
{
	return d_background_image->pixmap();
}

QSize TDisplayWidget::Label_Size(void) const 
{
	return d_label_item_size - QSize(PIXMAP_BORDER_SIZE,PIXMAP_BORDER_SIZE);
}

void TDisplayWidget::Change_Scale(
	const double						&scale)
{
	this->resetTransform();
	this->scale(scale,scale);
}

void TDisplayWidget::Set_Label_Size(
	const QSize							&size,
	const int							font_size)
{
	QPointF								target;
	QPointF								offset;
	int									border_type;
	int									value;
	QString								text;
	bool								is_selected;
	std::vector<QGraphicsPixmapItem*>::iterator item_iter;
	
	d_label_item_size = size + QSize(PIXMAP_BORDER_SIZE,PIXMAP_BORDER_SIZE);
	d_font_size = font_size;
	
	// update item pixmap at default size with RGBA
	QImage								item(d_label_item_size,QImage::Format_ARGB32);
	
	item.fill(Qt::transparent);
	d_item_pixmap = QPixmap::fromImage(item);
	
	for(item_iter = d_label_item_pixmaps.begin();item_iter != d_label_item_pixmaps.end();++item_iter)
	{
		value = (*item_iter)->data(ITEM_ID_VALUE_FIELD).toInt();
		text = (*item_iter)->data(ITEM_ID_TEXT_FIELD).toString();
		border_type = static_cast<int>((*item_iter)->data(ITEM_BORDER_TYPE_FIELD).toInt());
		is_selected = (0 != (*item_iter)->data(ITEM_SELECTED_FIELD).toInt());

		if(text.length())
		{
			(*item_iter)->setPixmap(this->Render_Pixmap(text,border_type,is_selected));
		}
		else
		{
			(*item_iter)->setPixmap(this->Render_Pixmap(value,border_type,is_selected));
		}
		
		target = (*item_iter)->data(ITEM_POS_FIELD).toPointF();
		offset = QPointF(d_label_item_size.width()/2,d_label_item_size.height()/2);
		
//		center_target.setX(center_target.x() - center_offset.x());
//		center_target.setY(center_target.y() - center_offset.y());
		
		target -= offset;

		(*item_iter)->setPos(target);
	}
}

void TDisplayWidget::Set_Label_Color(
	const QColor						&color)
{
//	QPointF								target;
//	QPointF								offset;
	int									border_type;
	int									value;
	QString								text;
	bool								is_selected;
	std::vector<QGraphicsPixmapItem*>::iterator item_iter;
	
	d_label_color = color;
	
	// update item pixmap at default size with RGBA
	QImage								item(d_label_item_size,QImage::Format_ARGB32);
	
	item.fill(Qt::transparent);
	d_item_pixmap = QPixmap::fromImage(item);
	
	for(item_iter = d_label_item_pixmaps.begin();item_iter != d_label_item_pixmaps.end();++item_iter)
	{
		value = (*item_iter)->data(ITEM_ID_VALUE_FIELD).toInt();
		text = (*item_iter)->data(ITEM_ID_TEXT_FIELD).toString();
		border_type = static_cast<int>((*item_iter)->data(ITEM_BORDER_TYPE_FIELD).toInt());
		is_selected = (0 != (*item_iter)->data(ITEM_SELECTED_FIELD).toInt());
		
		if(text.length())
		{
			(*item_iter)->setPixmap(this->Render_Pixmap(text,border_type,is_selected));
		}
		else
		{
			(*item_iter)->setPixmap(this->Render_Pixmap(value,border_type,is_selected));
		}
		
//		target = (*item_iter)->data(ITEM_POS_FIELD).toPointF();
//		offset = QPointF(d_label_item_size.width()/2,d_label_item_size.height()/2);
		
//		center_target.setX(center_target.x() - center_offset.x());
//		center_target.setY(center_target.y() - center_offset.y());
		
//		target -= offset;
		
//		(*item_iter)->setPos(target);
	}
}

void TDisplayWidget::Set_Selected_Items(
	const std::vector<int>				&items,
	const bool							allow_center_on_select)
{
	std::vector<int>::const_iterator	sel_iter;
	std::vector<QGraphicsPixmapItem*>::iterator item_iter;
	int									border_type;
	int									id;
	int									value;
	QString								text;
	bool								is_selected;
	bool								iter_item_selected;
	QPointF								center_target;
	bool								center_target_found(false);

	for(item_iter = d_label_item_pixmaps.begin();item_iter != d_label_item_pixmaps.end();++item_iter)
	{
		id = (*item_iter)->data(ITEM_ID_VALUE).toInt();
		border_type = (*item_iter)->data(ITEM_BORDER_TYPE_FIELD).toInt();
	
		is_selected = false;
		for(sel_iter = items.begin();sel_iter != items.end();sel_iter++)
		{
			if((*sel_iter) == id)
			{
				is_selected = true;
				break;
			}
		}
		
		iter_item_selected = (0 != (*item_iter)->data(ITEM_SELECTED_FIELD).toInt());
		
		if(iter_item_selected != is_selected)
		{
			value = (*item_iter)->data(ITEM_ID_VALUE_FIELD).toInt();
			text = (*item_iter)->data(ITEM_ID_TEXT_FIELD).toString();
			
			if(text.length())
			{
				(*item_iter)->setPixmap(this->Render_Pixmap(text,border_type,is_selected));
			}
			else
			{
				(*item_iter)->setPixmap(this->Render_Pixmap(value,border_type,is_selected));
			}
			
			(*item_iter)->setData(ITEM_SELECTED_FIELD,is_selected ? 1:0);
		}
		
		if(is_selected)
		{
			center_target = (*item_iter)->data(ITEM_POS_FIELD).toPointF();
			center_target_found = true;
		}
	}
	
	if(allow_center_on_select)
	{
		if(center_target_found && d_track_label)
		{
			this->centerOn(center_target);
		}
	}
}

void TDisplayWidget::Clear(void)
{
	std::vector<QGraphicsPixmapItem*>::iterator item_iter;

	for(item_iter = d_label_item_pixmaps.begin();item_iter != d_label_item_pixmaps.end();++item_iter)
	{
		d_graphics_scene->removeItem(*item_iter);
		delete (*item_iter);
	}
	
	d_label_item_pixmaps.clear();
	
	d_background_image->setPixmap(QPixmap(":/images/default.png"));
	d_background_image->setPos(0,0);
	
	this->Update_Scene_Rect();
}

void TDisplayWidget::Remove_Items(void)
{
	std::vector<QGraphicsPixmapItem*>::iterator item_iter;
	
	for(item_iter = d_label_item_pixmaps.begin();item_iter != d_label_item_pixmaps.end();++item_iter)
	{
		d_graphics_scene->removeItem(*item_iter);
		delete (*item_iter);
	}
	
	d_label_item_pixmaps.clear();
}

void TDisplayWidget::Remove_Item(
	const int							id)
{
	std::vector<QGraphicsPixmapItem*>::iterator item_iter;
	int									item_id;
	
	for(item_iter = d_label_item_pixmaps.begin();item_iter != d_label_item_pixmaps.end();++item_iter)
	{
		item_id = (*item_iter)->data(ITEM_ID_VALUE).toInt();
		
		if(item_id == id)
		{
			d_graphics_scene->removeItem(*item_iter);
			delete (*item_iter);
			
			d_label_item_pixmaps.erase(item_iter);
			
			return;
		}
	}
}

void TDisplayWidget::Set_Background_Image(
	const QPixmap						&pixmap)
{
	d_background_image->setPixmap(pixmap);
	d_background_image->setPos(0,0);
	
	this->Update_Scene_Rect();
}

QGraphicsPixmapItem* TDisplayWidget::Add_Item(
	const int							id,
	const int							value,
	const QString						&text,
	const int							border_type)
{
	QGraphicsPixmapItem					*graphics_item;
	QPoint								point;
	QPointF								center_offset;
	QPointF								center_target;
	QPointF								upper_left_target;
	
	if(this->Item_Exists(id))
	{
		return 0;
	}
	
	point = this->mapFromScene(QPointF(0,0));
	
	if(point.x() < 0 || point.y() < 0)
	{
		upper_left_target = this->mapToScene(QPoint(0,0));
	}
	else
	{
		upper_left_target = QPointF(0,0);
	}
							 
	center_offset = QPointF(d_label_item_size.width()/2,d_label_item_size.height()/2);
	
	center_target = upper_left_target;
	center_target.setX(center_target.x() + center_offset.x());
	center_target.setY(center_target.y() + center_offset.y());
	
	if(text.length())
	{
		graphics_item = d_graphics_scene->addPixmap(this->Render_Pixmap(text,border_type,false));
	}
	else
	{
		graphics_item = d_graphics_scene->addPixmap(this->Render_Pixmap(value,border_type,false));
	}
	
	graphics_item->setTransformationMode(Qt::SmoothTransformation);

	if(graphics_item)
	{		
		graphics_item->setPos(upper_left_target);
		graphics_item->setZValue(1);	// make sure it is drawn on top;
		
		graphics_item->setData(ITEM_ID_VALUE,id);
		graphics_item->setData(ITEM_ID_VALUE_FIELD,value);
		graphics_item->setData(ITEM_ID_TEXT_FIELD,text);
		graphics_item->setData(ITEM_POS_FIELD,center_target);
		graphics_item->setData(ITEM_BORDER_TYPE_FIELD,border_type);
		graphics_item->setData(ITEM_SELECTED_FIELD,0);

		d_label_item_pixmaps.push_back(graphics_item);
				
		return graphics_item;
	}

    return 0;
}

void TDisplayWidget::Move_Item(
	const int							id,
	const QPointF						&pos)
{
	std::vector<QGraphicsPixmapItem*>::iterator item_iter;
	QPointF								upper_left;
	int									item_id;
	
	for(item_iter = d_label_item_pixmaps.begin();item_iter != d_label_item_pixmaps.end();++item_iter)
	{
		item_id = (*item_iter)->data(ITEM_ID_VALUE).toInt();
		
		if(id == item_id)
		{
			upper_left.setX(pos.x() - d_label_item_size.width()/2);
			upper_left.setY(pos.y() - d_label_item_size.height()/2);
			
			(*item_iter)->setPos(upper_left);
			(*item_iter)->setData(ITEM_POS_FIELD,pos);
			
			emit Update_Item_Position(id,pos);
			
			return;
		}
	}
}

void TDisplayWidget::Change_Item_Border(
	const int							id,
	const int							border_type)
{
	std::vector<QGraphicsPixmapItem*>::iterator item_iter;
	int									item_id;
	bool								is_selected;
	int									value;
	QString								text;
	
	for(item_iter = d_label_item_pixmaps.begin();item_iter != d_label_item_pixmaps.end();++item_iter)
	{
		item_id = (*item_iter)->data(ITEM_ID_VALUE).toInt();
		
		if(item_id == id)
		{
			(*item_iter)->setData(ITEM_BORDER_TYPE_FIELD,static_cast<int>(border_type));
			is_selected = (0 != (*item_iter)->data(ITEM_SELECTED_FIELD).toInt());
			
			value = (*item_iter)->data(ITEM_ID_VALUE_FIELD).toInt();
			text = (*item_iter)->data(ITEM_ID_TEXT_FIELD).toString();
		
			if(text.length())
			{
				(*item_iter)->setPixmap(this->Render_Pixmap(text,border_type,is_selected));
			}
			else
			{
				(*item_iter)->setPixmap(this->Render_Pixmap(value,border_type,is_selected));
			}
			
			return;
		}
	}
}

void TDisplayWidget::Change_Item_Value(
	const int							id, 
	const int							value)
{
	std::vector<QGraphicsPixmapItem*>::iterator item_iter;
	int									item_id;
	int									border_type;
	bool								is_selected;
	QString								text;
	
	for(item_iter = d_label_item_pixmaps.begin();item_iter != d_label_item_pixmaps.end();++item_iter)
	{
		item_id = (*item_iter)->data(ITEM_ID_VALUE).toInt();
		
		if(item_id == id)
		{
			(*item_iter)->setData(ITEM_ID_VALUE_FIELD,value);
			text = (*item_iter)->data(ITEM_ID_TEXT_FIELD).toString();
			
			border_type = static_cast<int>((*item_iter)->data(ITEM_BORDER_TYPE_FIELD).toInt());
			is_selected = (0 != (*item_iter)->data(ITEM_SELECTED_FIELD).toInt());

			if(text.length())
			{
				(*item_iter)->setPixmap(this->Render_Pixmap(text,border_type,is_selected));
			}
			else
			{
				(*item_iter)->setPixmap(this->Render_Pixmap(value,border_type,is_selected));
			}
			
			return;
		}
	}
}

void TDisplayWidget::Change_Item_Text(
	const int							id,
	const QString						&text)
{
	std::vector<QGraphicsPixmapItem*>::iterator item_iter;
	int									item_id;
	int									border_type;
	bool								is_selected;
	int									value;
	
	for(item_iter = d_label_item_pixmaps.begin();item_iter != d_label_item_pixmaps.end();++item_iter)
	{
		item_id = (*item_iter)->data(ITEM_ID_VALUE).toInt();
		
		if(item_id == id)
		{
			(*item_iter)->setData(ITEM_ID_TEXT_FIELD,text);
			value = (*item_iter)->data(ITEM_ID_VALUE_FIELD).toInt();

			border_type = static_cast<int>((*item_iter)->data(ITEM_BORDER_TYPE_FIELD).toInt());
			is_selected = (0 != (*item_iter)->data(ITEM_SELECTED_FIELD).toInt());
			
			if(text.length())
			{
				(*item_iter)->setPixmap(this->Render_Pixmap(text,border_type,is_selected));
			}
			else
			{
				(*item_iter)->setPixmap(this->Render_Pixmap(value,border_type,is_selected));
			}
			
			return;
		}
	}
}

void TDisplayWidget::mouseDoubleClickEvent(
	QMouseEvent							*event)
{
	Q_UNUSED(event);
}

void TDisplayWidget::mouseMoveEvent(
	QMouseEvent							*event)
{
	Qt::MouseButtons					buttons;
	QPointF								point;
	QPointF								offset;

	buttons = event->buttons();
	
	if((buttons & Qt::LeftButton) == Qt::LeftButton && d_moving_item != 0)
	{
		point = this->mapToScene(event->pos());
		
		offset = point - d_moving_item_last_position;
		
		if(offset.manhattanLength() > 1)
		{
			d_moving_item_last_position = point;
			
			point = d_moving_item->data(ITEM_POS_FIELD).toPointF();
			point += offset;
			d_moving_item->setData(ITEM_POS_FIELD,point);
			
			d_moving_item->moveBy(offset.x(),offset.y());
			
			d_moving_item_moved = true;
		}
	}
	else if(d_moving_item == 0)
	{
		QGraphicsView::mouseMoveEvent(event);
	}
}

void TDisplayWidget::mousePressEvent(
	QMouseEvent							*event)
{
	QPointF								point;
	QGraphicsItem						*graphics_item;
	int									item_id;
	
	if((d_mouse_down_buttons & Qt::LeftButton) == Qt::LeftButton)
	{
		this->setDragMode(QGraphicsView::ScrollHandDrag);
	}

	QGraphicsView::mousePressEvent(event);

	d_mouse_down_buttons = event->buttons();
	point = this->mapToScene(event->pos());
	
	d_moving_item_moved = false;
	d_moving_item = 0;
	
	if((d_mouse_down_buttons & Qt::LeftButton) == Qt::LeftButton)
	{

		graphics_item = d_graphics_scene->itemAt(point,this->transform());
		
		if(graphics_item != 0 && graphics_item != d_background_image)
		{			
			d_moving_item = reinterpret_cast<QGraphicsPixmapItem*>(graphics_item);
			
			if(d_moving_item)
			{
				d_moving_item_last_position = point;
				item_id = d_moving_item->data(ITEM_ID_VALUE).toInt();
												
				emit Item_Clicked(item_id);
			}
		}
	}
	
	if((d_mouse_down_buttons & Qt::RightButton) == Qt::RightButton)
	{
		emit Right_Mouse_Click(point);
	}
}

void TDisplayWidget::mouseReleaseEvent(
	QMouseEvent							*event)
{
	Qt::MouseButtons					buttons;
	QPointF								point;
	int									id;
	
	QGraphicsView::mouseReleaseEvent(event);
	
	this->setDragMode(QGraphicsView::NoDrag);


	if((d_mouse_down_buttons & Qt::LeftButton) == Qt::LeftButton)
	{
		if(d_moving_item)
		{
			point = d_moving_item->data(ITEM_POS_FIELD).toPointF();
			id = d_moving_item->data(ITEM_ID_VALUE).toInt();

			emit Update_Item_Position(id,point);
		}
		else if(!d_moving_item_moved)
		{
			emit Item_Clicked(-1);
		}

		d_moving_item = 0;
	}	
}

void TDisplayWidget::Update_Scene_Rect(void)
{
	QRectF								bounding_rect;
	
	bounding_rect = d_graphics_scene->itemsBoundingRect();
	
	d_graphics_scene->setSceneRect(bounding_rect);
}

QPixmap TDisplayWidget::Render_Pixmap(
	const int							value,
	const int							&border_type,
	const bool							is_selected) const
{
	return this->Render_Pixmap(QString("%1").arg(value),border_type,is_selected);
}

QPixmap TDisplayWidget::Render_Pixmap(
	const QString						&text,
	const int							&border_type,
	const bool							is_selected) const
{
	QPixmap								item(d_item_pixmap);
	QColor								background_color(QColor(255,255,255,1));	// alpha cannot be zero otherwise selection difficult
	QBrush								brush(background_color);
	QPen								pen(d_label_color);
	QPen								selected_pen(QColor(0,0,255,128));
	QBrush								selected_brush(QColor(0,0,255,64));
	QFont								font("Arial",d_font_size,QFont::Bold);
	QPointF								points[5];
	QSize								draw_size(d_label_item_size - QSize(PIXMAP_BORDER_SIZE,PIXMAP_BORDER_SIZE));
	QSize								selected_draw_size(d_label_item_size - QSize(1,1));
	int									draw_radius(draw_size.width()/2);
	int									selected_draw_radius(selected_draw_size.width()/2);
	QRect								text_bounding_rect;
	double								scale;
	QString								label_text;
	QPainter							painter(&item);

	pen.setWidth(2);
	
	painter.setBrush(brush);
	painter.setFont(font);
	painter.setRenderHint(QPainter::Antialiasing,true);
	
	painter.fillRect(item.rect(), brush);

	painter.translate(d_label_item_size.width()/2,d_label_item_size.height()/2);
	
	if(is_selected)
	{
		painter.setPen(selected_pen);
		painter.setBrush(selected_brush);
		painter.drawRect(-selected_draw_radius,-selected_draw_radius,selected_draw_size.width()-1,selected_draw_size.height()-1);
	}
			
	painter.setPen(pen);
	painter.setBrush(brush);

	switch(border_type)
	{
		case TBorderType::BORDER_TYPE_ELLIPSE:
			painter.drawEllipse(-draw_radius,-draw_radius,draw_size.width()-1,draw_size.height()-1);
			break;
			
		case TBorderType::BORDER_TYPE_SQUARE:
			painter.drawRect(-draw_radius,-draw_radius,draw_size.width()-1,draw_size.height()-1);
			break;
			
		case TBorderType::BORDER_TYPE_TRIANGLE:
			
			points[0] = QPointF(0,-draw_radius);
			points[1] = QPointF(-draw_radius,0);
			points[2] = QPointF(-draw_radius,draw_radius);
			points[3] = QPointF(draw_radius,draw_radius);
			points[4] = QPointF(draw_radius,0); 
			
			painter.drawPolygon(points,5);
			break;
			
		case TBorderType::BORDER_TYPE_DIAMOND:
			
			points[0] = QPointF(0,-draw_radius);
			points[1] = QPointF(-draw_radius,0);
			points[2] = QPointF(0,draw_radius);
			points[3] = QPointF(draw_radius,0);
			
			painter.drawPolygon(points,4);
			break;
	
		case TBorderType::BORDER_TYPE_NONE:
		default:
			break;
	}
	
	label_text = text;
	text_bounding_rect = painter.boundingRect(0,0,0,0,0,label_text);
	
	// 0.67 is for minimum width of each side to text
	scale = 0.67 * static_cast<double>(draw_size.width())/static_cast<double>(text_bounding_rect.width());
	
	// text won't fit inside area.  Display '*'
	if(scale < 0.2)
	{
		scale = 1.0;
		label_text = "*";
	}

	if(scale < 1.0)
	{
		font.setPointSizeF(static_cast<double>(d_font_size) * scale);
		painter.setFont(font);
	}
	
	if(border_type == TBorderType::BORDER_TYPE_TRIANGLE)
	{
		painter.drawText(-draw_radius,0,draw_size.width(),draw_size.height() / 2,Qt::AlignVCenter | Qt::AlignHCenter,label_text);
	
	}
	else
	{
		painter.drawText(-draw_radius,-draw_radius,draw_size.width(),draw_size.height(),Qt::AlignVCenter | Qt::AlignHCenter,label_text);
	}
	
	return item;
}


