/////////////////////////////////////////////////////////////////////
//
//            X   X           X
//           XXX XX         XX
//          XXXXXXXX      XXX
//         XX X XXXXXXXXXXX
//        XXXXX XXXXXXXXX
//       XXXXX XXXXXXXXXX
//            XXX XXX XXX
//           XXX XX   XX
//           X   X     X
//
//    Copyright (C) 2003-2020  Ron Jakl
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
/////////////////////////////////////////////////////////////////////


#include <QApplication>
#include <QGridLayout>
#include <QHBoxLayout>
#include <QListWidget>
#include <QListWidgetItem>
#include <QSpacerItem>
#include <QToolButton>
#include <QListWidgetItem>
#include <QLineEdit>
#include <assert.h>

#include "pagewidget.h"

TPageWidget::TPageWidget(
	const QWidget						*parent,
	const Qt::WindowFlags				flags)
:QWidget(const_cast<QWidget*>(parent),flags)
{
	QGridLayout							*widget_layout;
	QHBoxLayout							*button_hlayout;
	QSpacerItem							*button_hspacer;

	this->resize(280,250);

	widget_layout = new QGridLayout(this);

	d_page_list = new QListWidget(this);
	d_page_list->setSelectionMode(QAbstractItemView::ExtendedSelection);
	d_page_list->setDragDropMode(QAbstractItemView::InternalMove);
	widget_layout->addWidget(d_page_list,0,0,1,1);

	button_hlayout = new QHBoxLayout();
	
	d_page_name_edit = new QLineEdit(this);
	d_page_name_edit->setAlignment(Qt::AlignCenter);
	button_hlayout->addWidget(d_page_name_edit);

	button_hspacer = new QSpacerItem(0,0,QSizePolicy::Expanding,QSizePolicy::Minimum);
	button_hlayout->addItem(button_hspacer);

	d_add_page_button = new QToolButton(this);
	button_hlayout->addWidget(d_add_page_button);

	d_remove_page_button = new QToolButton(this);
	button_hlayout->addWidget(d_remove_page_button);
	widget_layout->addLayout(button_hlayout,1,0,1,1);
	
	this->setWindowTitle(QStringLiteral("Page List"));
	
	d_add_page_button->setToolTip(QStringLiteral("Add Page"));
	d_remove_page_button->setToolTip(QStringLiteral("Remove Page"));
	d_page_name_edit->setToolTip(QStringLiteral("Page Item Name"));

	d_add_page_button->setText(QStringLiteral("+"));
	d_remove_page_button->setText(QStringLiteral("-"));
	
	// defaults
	d_remove_page_button->setEnabled(false);
	d_page_name_edit->setEnabled(false);
	
	connect(d_page_list,SIGNAL(itemSelectionChanged(void)),this,SLOT(Page_Selection_Changed(void)));
	connect(d_add_page_button,SIGNAL(clicked(void)),this,SIGNAL(Add_Page(void)));
	connect(d_remove_page_button,SIGNAL(clicked(void)),this,SLOT(Remove_Page(void)));
	connect(d_page_name_edit,SIGNAL(textEdited(const QString&)),this,SLOT(Page_Name_Edited(const QString&)));
}

TPageWidget::~TPageWidget(void)
{
}

QList<int> TPageWidget::Page_Ids(void) const
{
	QList<int>							list;
	QListWidgetItem						*item;
	int									cntr;
	int									id;
	
	for(cntr = 0;cntr < d_page_list->count();++cntr)
	{
		item = d_page_list->item(cntr);
		id = item->data(Qt::UserRole).toInt();
				
		list.push_back(id);
	}
	
	return list;
}

QString TPageWidget::Page_Name(
	const int							page_id) const
{
	QListWidgetItem						*item;
	int									cntr;
	int									id;
	
	for(cntr = 0;cntr < d_page_list->count();++cntr)
	{
		item = d_page_list->item(cntr);
		id = item->data(Qt::UserRole).toInt();
		
		if(id == page_id)
		{
			return item->text();
		}
	}
	
	return QString();
}

void TPageWidget::Clear(void)
{
	d_page_list->clear();
}

void TPageWidget::Add_Page(
	const int							page_id)
{
	QListWidgetItem						*item;
	
	item = new QListWidgetItem(d_page_list);
	
	item->setData(Qt::UserRole,page_id);
	item->setText(QString("Page %1").arg(page_id));
}

void TPageWidget::Remove_Page(
	const int							page_id)
{
	QListWidgetItem						*item;
	int									cntr;
	int									id;
	
	for(cntr = 0;cntr < d_page_list->count();++cntr)
	{
		item = d_page_list->item(cntr);
		id = item->data(Qt::UserRole).toInt();
		
		if(id == page_id)
		{
			item = d_page_list->takeItem(cntr);
			delete item;
			
			return;
		}
	}
}

void TPageWidget::Set_Active_Page(
	const int							page_id)
{
	QListWidgetItem						*item;
	int									cntr;
	int									id;
	
	for(cntr = 0;cntr < d_page_list->count();++cntr)
	{
		item = d_page_list->item(cntr);
		id = item->data(Qt::UserRole).toInt();
		
		if(id == page_id)
		{
			d_page_list->setCurrentItem(item,QItemSelectionModel::ClearAndSelect);
			return;
		}
	}
}
void TPageWidget::Set_Page_Name(
	const int							page_id,
	const QString						&name)
{
	QListWidgetItem						*item;
	int									cntr;
	int									id;
	
	for(cntr = 0;cntr < d_page_list->count();++cntr)
	{
		item = d_page_list->item(cntr);
		id = item->data(Qt::UserRole).toInt();
		
		if(id == page_id)
		{
			item->setText(name);
			return;
		}
	}
}

void TPageWidget::Page_Name_Edited(
	const QString						&text)
{
	QList<QListWidgetItem*>				sel_items;
	
	sel_items = d_page_list->selectedItems();

	if(sel_items.size() == 1)
	{
		sel_items[0]->setText(text);
	}
}

void TPageWidget::Page_Selection_Changed(void)
{
	QList<QListWidgetItem*>				sel_items;
	int									page_id;
	
	sel_items = d_page_list->selectedItems();
	
	d_remove_page_button->setEnabled(sel_items.size());
	d_page_name_edit->setEnabled(sel_items.size() == 1);

	if(sel_items.size() == 1)
	{
		d_page_name_edit->setText(sel_items[0]->text());
		
		page_id = sel_items[0]->data(Qt::UserRole).toInt();
		emit Selection_Changed(page_id);
	}
	else
	{
		d_page_name_edit->setText(QString());

		emit Selection_Changed(-1);
	}
	
}

void TPageWidget::Remove_Page(void)
{
	QList<QListWidgetItem*>				sel_items;
	QList<QListWidgetItem*>::iterator	iter;
	QList<int>							page_ids;

	sel_items = d_page_list->selectedItems();
	
	for(iter = sel_items.begin();iter != sel_items.end();++iter)
	{
		page_ids.push_back((*iter)->data(Qt::UserRole).toInt());
	}
	
	emit Remove_Pages(page_ids);
}


