/////////////////////////////////////////////////////////////////////
//
//            X   X           X
//           XXX XX         XX
//          XXXXXXXX      XXX
//         XX X XXXXXXXXXXX
//        XXXXX XXXXXXXXX
//       XXXXX XXXXXXXXXX
//            XXX XXX XXX
//           XXX XX   XX
//           X   X     X
//
//    Copyright (C) 2003-2020  Ron Jakl
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
/////////////////////////////////////////////////////////////////////


#include <QAction>
#include <QApplication>
#include <QGuiApplication>
#include <QDockWidget>
#include <QMenu>
#include <QMenuBar>
#include <QSize>
#include <QToolBar>
#include <QStatusBar>
#include <QWidget>
#include <QSlider>
#include <QLabel>
#include <QHBoxLayout>
#include <QVBoxLayout>
#include <QSizePolicy>
#include <QColor>
#include <QRect>
#include <QPrinter>
#include <QPainter>
#include <QCloseEvent>
#include <QFileDialog>
#include <QFile>
#include <QImage>
#include <QList>
#include <QDateTime>
#include <QBuffer>
#include <QSettings>
#include <QScreen>
#include <QFont>
#include <QFontMetrics>
#include <cmath>
#include <time.h>
#include <assert.h>

#include "../../core/messagebox.h"
#include "../../core/xmlfile.h"
#include "../../core/conversion.h"

#include "labelwidget.h"
#include "settingswidget.h"
#include "pagewidget.h"
#include "displaywidget.h"
#include "reindexdialog.h"
#include "reportheaderdialog.h"
#include "data.h"

#include "reportroadmap.h"

TReportRoadMap::TReportRoadMap(
	const QWidget						*parent,
	const Qt::WindowFlags				flags)
:QMainWindow(const_cast<QWidget*>(parent),flags)
{
	QMenu								*edit_menu;
	QMenu								*file_menu;
	QMenu								*help_menu;
	QMenu								*view_menu;
	QMenuBar							*menu;
	QToolBar							*toolbar;
	QStatusBar							*statusbar;
	QLabel								*zoom_label;
	QLabel								*copyright_label;
	QWidget								*zoom_widget;
	QWidget								*central_widget;
	QVBoxLayout							*central_widget_layout;
	QHBoxLayout							*zoom_hlayout;
	QString								data_path;
	QDir								data_location;
	QPoint								position;
	QSize								size;
	QRect								desktop_rect;
	QRect								window_rect;
	bool								state;
	QColor								color;
	int									ival;
	QFont								copyright_font;

	this->resize(800,600);
	this->setIconSize(QSize(32,32));
	
	d_reindex_dialog = 0;
	d_report_header_dialog = 0;
	
	copyright_font.setFamily(QStringLiteral("Helvetica"));
	copyright_font.setPointSize(8);
	copyright_font.setItalic(true);
	
	data_location = QDir::home();
	
	if(!data_location.exists(".reportroadmap"))
	{
		data_location.mkdir(".reportroadmap");
	}
	
	data_location.cd(".reportroadmap");
	
	data_path = data_location.absolutePath();

	d_settings = new QSettings(data_path + "/Settings.ini",QSettings::IniFormat,this);
	
	position = d_settings->value("Mainwindow_Position", QPoint(30,30)).toPoint();
	size = d_settings->value("Mainwindow_Size", QSize(800, 600)).toSize();
	
	desktop_rect = QGuiApplication::primaryScreen()->availableGeometry();
	
	window_rect.moveTopLeft(position);
	window_rect.setSize(size);
	
	window_rect = window_rect & desktop_rect;
	
	if(window_rect.isValid())
	{
		this->resize(window_rect.size());
		this->move(window_rect.topLeft());
	}
	else
	{
		this->resize(800,600);
		this->move(QPoint(30,30));
	}

	d_msg_box = new TMessageBox(this);

	d_action_new  = new QAction(this);
	d_action_new->setShortcut(QKeySequence::New);
	d_action_new->setIcon(QIcon(":/menu/file_new.png"));

	d_action_open = new QAction(this);
	d_action_open->setShortcut(QKeySequence::Open);
	d_action_open->setIcon(QIcon(":/menu/file_open.png"));

	d_action_save = new QAction(this);
	d_action_save->setShortcut(QKeySequence::Save);
	d_action_save->setIcon(QIcon(":/menu/file_save.png"));
	d_action_save->setEnabled(false);

	d_action_saveas = new QAction(this);
	d_action_saveas->setShortcut(QKeySequence::SaveAs);

	d_action_import = new QAction(this);
	d_action_import->setWhatsThis(QStringLiteral("Import background image for road map."));
	d_action_import->setIcon(QIcon(":/menu/file_import.png"));

	d_action_export_images = new QAction(this);
	d_action_export_images->setWhatsThis(QStringLiteral("Export road map page images."));
	d_action_export_images->setIcon(QIcon(":/menu/file_export_image.png"));
	d_action_export_images->setEnabled(false);

	d_action_export_pdf = new QAction(this);
	d_action_export_pdf->setWhatsThis(QStringLiteral("Export road map pages as PDF."));
	d_action_export_pdf->setIcon(QIcon(":/menu/file_export_pdf.png"));
	d_action_export_pdf->setEnabled(false);
	
	d_action_quit = new QAction(this);
	d_action_quit->setShortcut(QKeySequence::Quit);

	d_action_reindex = new QAction(this);
	d_action_reindex->setText(QStringLiteral("&Reindex..."));
	d_action_reindex->setEnabled(false);
	
	d_action_about = new QAction(this);

	menu = new QMenuBar(this);
	this->setMenuBar(menu);

	file_menu = new QMenu(menu);
	file_menu->addAction(d_action_new);
	file_menu->addSeparator();
	file_menu->addAction(d_action_open);
	file_menu->addAction(d_action_save);
	file_menu->addAction(d_action_saveas);
	file_menu->addSeparator();
	file_menu->addAction(d_action_import);
	file_menu->addAction(d_action_export_images);
	file_menu->addAction(d_action_export_pdf);
	file_menu->addSeparator();
	file_menu->addAction(d_action_quit);
	menu->addMenu(file_menu);

	view_menu = new QMenu(menu);
	menu->addMenu(view_menu);

	edit_menu = new QMenu(menu);
	edit_menu->addAction(d_action_reindex);
	menu->addMenu(edit_menu);

	help_menu = new QMenu(menu);
	help_menu->addAction(d_action_about);
	menu->addMenu(help_menu);

	menu->resize(800,22);
	
	menu->addAction(file_menu->menuAction());
	menu->addAction(edit_menu->menuAction());
	menu->addAction(view_menu->menuAction());
	menu->addAction(help_menu->menuAction());

	toolbar = new QToolBar(this);
	this->addToolBar(Qt::TopToolBarArea,toolbar);
	
	toolbar->addAction(d_action_new);
	toolbar->addSeparator();
	toolbar->addAction(d_action_open);
	toolbar->addAction(d_action_save);
	toolbar->addSeparator();
	toolbar->addAction(d_action_import);
	toolbar->addAction(d_action_export_images);
	toolbar->addAction(d_action_export_pdf);

	statusbar = new QStatusBar(this);
	this->setStatusBar(statusbar);
	
	zoom_widget = new QWidget(this);
	
	zoom_hlayout = new QHBoxLayout(zoom_widget);
	
	zoom_label = new QLabel(zoom_widget);
	zoom_hlayout->addWidget(zoom_label);
	
	d_zoom_slider = new QSlider(zoom_widget);
	d_zoom_slider->setOrientation(Qt::Horizontal);
	d_zoom_slider->setRange(-100,100);
	d_zoom_slider->setValue(0);
	zoom_hlayout->addWidget(d_zoom_slider);
	
	d_zoom_scale_label = new QLabel(zoom_widget);
	zoom_hlayout->addWidget(d_zoom_scale_label);
	
	statusbar->addPermanentWidget(zoom_widget);

	d_label_dock = new QDockWidget(this);
	d_label_dock->setFeatures(QDockWidget::AllDockWidgetFeatures);
	d_label_dock->setAllowedAreas(Qt::LeftDockWidgetArea|Qt::RightDockWidgetArea);
	view_menu->addAction(d_label_dock->toggleViewAction());
	this->addDockWidget(Qt::LeftDockWidgetArea,d_label_dock);
	
	d_page_dock = new QDockWidget(this);
	d_page_dock->setFeatures(QDockWidget::AllDockWidgetFeatures);
	d_page_dock->setAllowedAreas(Qt::LeftDockWidgetArea|Qt::RightDockWidgetArea);
	view_menu->addAction(d_page_dock->toggleViewAction());
	this->addDockWidget(Qt::RightDockWidgetArea,d_page_dock);

	d_settings_dock = new QDockWidget(this);
	d_settings_dock->setFeatures(QDockWidget::AllDockWidgetFeatures);
	d_settings_dock->setAllowedAreas(Qt::LeftDockWidgetArea|Qt::RightDockWidgetArea);
	view_menu->addAction(d_settings_dock->toggleViewAction());
	this->addDockWidget(Qt::RightDockWidgetArea,d_settings_dock);
	
// #ifdef Q_OS_MAC
// 	dock->setFeatures(QDockWidget::DockWidgetMovable);
// #else
// 	dock->setFeatures(QDockWidget::DockWidgetFloatable|QDockWidget::DockWidgetMovable);
// #endif

	d_label_widget = new TLabelWidget(this);
	d_label_dock->setWidget(d_label_widget);

	d_settings_widget = new TSettingsWidget(this);
	d_settings_dock->setWidget(d_settings_widget);

	d_page_widget = new TPageWidget(this);
	d_page_dock->setWidget(d_page_widget);
	
	central_widget = new QWidget(this);
	
	central_widget_layout = new QVBoxLayout(central_widget);
	central_widget_layout->setContentsMargins(0,0,0,-1);
	
	d_display_widget = new TDisplayWidget(central_widget);
	d_display_widget->setVerticalScrollBarPolicy(Qt::ScrollBarAsNeeded);
	d_display_widget->setHorizontalScrollBarPolicy(Qt::ScrollBarAsNeeded);
	central_widget_layout->addWidget(d_display_widget);
	
	copyright_label = new QLabel(central_widget);
	copyright_label->setFont(copyright_font);
	copyright_label->setAlignment(Qt::AlignLeft | Qt::AlignVCenter);
	central_widget_layout->addWidget(copyright_label);
	
	this->setCentralWidget(central_widget);
	
	this->setWindowTitle(QStringLiteral("Report Road Map"));	// change version in Help - About
	
	d_action_new->setToolTip(QStringLiteral("New"));
	d_action_open->setToolTip(QStringLiteral("Open"));
	d_action_save->setToolTip(QStringLiteral("Save"));
	d_action_saveas->setToolTip(QStringLiteral("Save As"));
	d_action_import->setToolTip(QStringLiteral("Import"));
	d_action_export_images->setToolTip(QStringLiteral("Export Images"));
	d_action_export_pdf->setToolTip(QStringLiteral("Export PDF"));
	d_zoom_slider->setToolTip(QStringLiteral("Image Magnification"));

	d_action_new->setText(QStringLiteral("New"));
	d_action_open->setText(QStringLiteral("Open ..."));
	d_action_save->setText(QStringLiteral("Save"));
	d_action_saveas->setText(QStringLiteral("Save As ..."));
	d_action_import->setText(QStringLiteral("Import ..."));
	d_action_export_images->setText(QStringLiteral("&Export Images ..."));
	d_action_export_pdf->setText(QStringLiteral("Export PDF ..."));
	d_action_quit->setText(QStringLiteral("Quit"));
	d_action_reindex->setText(QStringLiteral("Reindex Items ..."));
	d_action_about->setText(QStringLiteral("About ..."));
	zoom_label->setText(QStringLiteral("Zoom:"));
	d_zoom_scale_label->setText(QStringLiteral("1.00 X"));
	copyright_label->setText(QStringLiteral("  Copyright (C) 2020 Select Calibration Incorporated.  www.selectcalibration.ca"));

	file_menu->setTitle(QStringLiteral("File"));
	view_menu->setTitle(QStringLiteral("View"));
	edit_menu->setTitle(QStringLiteral("Edit"));
	help_menu->setTitle(QStringLiteral("Help"));
	
	d_label_dock->setWindowTitle(QStringLiteral("Item List"));
	d_settings_dock->setWindowTitle(QStringLiteral("Settings"));
	d_page_dock->setWindowTitle(QStringLiteral("Pages"));
	
	// settings
	d_working_path.setPath(d_settings->value("Working_Path",QDir::home().absolutePath()).toString());
	
	state = d_settings->value("Track_Selected_Label",true).toBool();
	d_display_widget->Set_Track_Label(state);
	d_settings_widget->Set_Track_Label(state);

	ival = d_settings->value("Label_Size",32).toInt();
	d_settings_widget->Set_Label_Size(ival);
	this->Label_Size_Changed(ival);
	
	d_settings_widget->Set_Display_Format(d_settings->value("Item_Display_Format",QStringLiteral("Item[%1]")).toString());
	
	color.setNamedColor(d_settings->value("Label_Color",QStringLiteral("#f79741")).toString());
	d_settings_widget->Set_Label_Color(color);
	d_display_widget->Set_Label_Color(color);
	
	d_settings_widget->Set_PDF_Orientation(d_settings->value("PDF_Orientation",0).toInt());
	d_settings_widget->Set_PDF_Resolution(d_settings->value("PDF_Resolution",2).toInt());
	
	d_label_widget->Set_Border_Type(d_settings->value("Border_Type",2).toInt());
	
	d_label_dock->setVisible(d_settings->value("Label_Visible",true).toBool());
	d_settings_dock->setVisible(d_settings->value("Settings_Visible",true).toBool());
	d_page_dock->setVisible(d_settings->value("Page_Visible",true).toBool());
	
	d_label_dock->setFloating(d_settings->value("Label_Floating",false).toBool());
	d_settings_dock->setFloating(d_settings->value("Settings_Floating",false).toBool());
	d_page_dock->setFloating(d_settings->value("Page_Floating",false).toBool());
	
	if(d_label_dock->isFloating()) d_label_dock->move(d_settings->value("Label_Position",QPoint(0,0)).toPoint());
	if(d_settings_dock->isFloating()) d_settings_dock->move(d_settings->value("Settings_Position",QPoint(0,0)).toPoint());
	if(d_page_dock->isFloating()) d_page_dock->move(d_settings->value("Page_Position",QPoint(0,0)).toPoint());

	
	
	// defaults
	d_block_label_tracking = false;
	d_data_changed = false;
	d_label_widget->setEnabled(false);

	connect(d_action_new,SIGNAL(triggered(void)), this, SLOT(File_New(void)));
	connect(d_action_open,SIGNAL(triggered(void)), this, SLOT(File_Open(void)));
	connect(d_action_import,SIGNAL(triggered(void)), this, SLOT(File_Import(void)));
	connect(d_action_save,SIGNAL(triggered(void)), this, SLOT(File_Save(void)));
	connect(d_action_saveas,SIGNAL(triggered(void)), this, SLOT(File_SaveAs(void)));
	connect(d_action_export_images,SIGNAL(triggered(void)), this, SLOT(File_Export_Images(void)));
	connect(d_action_export_pdf,SIGNAL(triggered(void)), this, SLOT(File_Export_Pdf(void)));
	connect(d_action_reindex,SIGNAL(triggered(void)), this, SLOT(Edit_Reindex(void)));
	connect(d_action_about,SIGNAL(triggered(void)), this, SLOT(Help_About(void)));

	connect(d_zoom_slider,SIGNAL(valueChanged(int)),this,SLOT(Image_Zoom_Changed(int)));

	connect(d_label_widget,SIGNAL(Add_Item(void)),this,SLOT(Add_Item(void)));
	connect(d_label_widget,SIGNAL(Remove_Item(void)),this,SLOT(Remove_Item(void)));
	connect(d_label_widget,SIGNAL(Selection_Changed(void)),this,SLOT(Item_Selection_Changed(void)));
	connect(d_label_widget,SIGNAL(Change_Item_Value(const int,const int)),this,SLOT(Change_Item_Value(const int,const int)));
	connect(d_label_widget,SIGNAL(Change_Item_Text(const int,const QString&)),this,SLOT(Change_Item_Text(const int,const QString&)));
	connect(d_label_widget,SIGNAL(Change_Item_Border(const int,int)),this,SLOT(Change_Item_Border(const int,const int)));
	
	connect(d_settings_widget,SIGNAL(Label_Size_Changed(int)),this,SLOT(Label_Size_Changed(int)));
	connect(d_settings_widget,SIGNAL(Color_Changed(const QColor)),this,SLOT(Color_Changed(const QColor)));
	connect(d_settings_widget,SIGNAL(Item_Display_Format_Changed(const QString&)),this,SLOT(Item_Display_Format_Changed(const QString&)));
	connect(d_settings_widget,SIGNAL(Track_Item_Option_Changed(bool)),this,SLOT(Track_Item_Option_Changed(bool)));
	
	connect(d_page_widget,SIGNAL(Selection_Changed(int)),this,SLOT(Page_Selection_Changed(int)));
	connect(d_page_widget,SIGNAL(Add_Page(void)),this,SLOT(Add_Page(void)));
	connect(d_page_widget,SIGNAL(Remove_Pages(const QList<int>&)),this,SLOT(Remove_Pages(const QList<int>&)));

	connect(d_display_widget,SIGNAL(Right_Mouse_Click(const QPointF&)),this,SLOT(Add_Item(const QPointF&)));
	connect(d_display_widget,SIGNAL(Item_Clicked(const int)),this,SLOT(Display_Item_Clicked(const int)));
	connect(d_display_widget,SIGNAL(Update_Item_Position(const int,const QPointF&)),this,SLOT(Update_Item_Position(const int,const QPointF&)));
}

TReportRoadMap::~TReportRoadMap(void)
{
}

void TReportRoadMap::File_New(void)
{
	if(d_data_changed)
	{
		d_msg_box->setText("Data Not Saved");
		d_msg_box->setInformativeText("Current data will be lost.  Do you want to continue?");
		d_msg_box->setDetailedText(QString());
		d_msg_box->setStandardButtons(QMessageBox::Yes | QMessageBox::No);
		d_msg_box->setDefaultButton(QMessageBox::No);
		d_msg_box->setIcon(QMessageBox::Question);
		
		if(d_msg_box->exec() == QMessageBox::No)
		{
			return;
		}
	}
	
	d_data.Clear();
	d_page_widget->Clear();
	d_label_widget->Clear();
	
	d_display_widget->Clear();
	d_display_widget->Set_Background_Image(QPixmap(":/images/background_default.png"));
	
	d_action_export_pdf->setEnabled(false);
	d_action_export_images->setEnabled(false);
	d_action_reindex->setEnabled(false);
	d_action_save->setEnabled(false);
	
	d_active_file_name = QString();
	d_data_changed = false;
}

void TReportRoadMap::File_Open(void)
{
	QString								file_name;
	int									index;

	if(d_data_changed)
	{
		d_msg_box->setText("Data Not Saved");
		d_msg_box->setInformativeText("Current data will be lost.  Do you want to continue?");
		d_msg_box->setDetailedText(QString());
		d_msg_box->setStandardButtons(QMessageBox::Yes | QMessageBox::No);
		d_msg_box->setDefaultButton(QMessageBox::No);
		d_msg_box->setIcon(QMessageBox::Question);
		
		if(d_msg_box->exec() == QMessageBox::No)
		{
			return;
		}
	}
	
	file_name = QFileDialog::getOpenFileName(this,
											 QStringLiteral("Select a ReportMap file to open"),
											 d_working_path.absolutePath(),
											 QStringLiteral("ReportMap files (*.reportmap *.rmd)"));
	
	if(file_name.length())
	{
		// legacy file check
		if(file_name.endsWith(".rmd",Qt::CaseInsensitive))
		{
			file_name.chop(4);
		
			if(!this->Load_Legacy_Data(file_name))
			{
				d_msg_box->setText("Error Loading Data");
				d_msg_box->setInformativeText("The legacy Report Road Map file cannot be loaded");
				d_msg_box->setDetailedText(d_last_error);
				d_msg_box->setStandardButtons(QMessageBox::Ok);
				d_msg_box->setDefaultButton(QMessageBox::Ok);
				d_msg_box->setIcon(QMessageBox::Critical);
				d_msg_box->exec();
				
			}
			
			d_active_file_name = QString();
		}
		else
		{
			if(!this->Load_Data(file_name))
			{
				d_msg_box->setText("Error Loading Data");
				d_msg_box->setInformativeText("The Report Road Map file cannot be loaded");
				d_msg_box->setDetailedText(d_last_error);
				d_msg_box->setStandardButtons(QMessageBox::Ok);
				d_msg_box->setDefaultButton(QMessageBox::Ok);
				d_msg_box->setIcon(QMessageBox::Critical);
				d_msg_box->exec();
				
			}
			
			d_active_file_name = file_name;
		}
		
		index = file_name.lastIndexOf('/');
		
		if(!(index < 0))
		{
			file_name.truncate(index);
		}
		
		d_working_path.setPath(file_name);
		d_data_changed = false;
	}
	
	d_action_export_pdf->setEnabled(d_data.Page_Count() > 0);
	d_action_export_images->setEnabled(d_data.Page_Count() > 0);
	d_action_reindex->setEnabled(d_data.Page_Count() > 0);
	d_action_save->setEnabled(d_data.Page_Count() > 0);
}

void TReportRoadMap::File_Import(void)
{
	QStringList							file_names;
	QStringList::const_iterator			iter;
	QString								text;
	int									index;
	bool								init(true);
	
	file_names = QFileDialog::getOpenFileNames(this,
											 QStringLiteral("Import one or more supported image files"),
											 d_working_path.absolutePath(),
											 QStringLiteral("Image Files (*.bmp *.gif *.jpg *.jpeg *.png *.pbm *.pgm *.ppm *.xbm *.xpm)"));
	
	
	for(iter = file_names.begin();iter != file_names.end();++iter)
	{
		this->Load_Image(*iter);
		
		if(init)
		{
			text = (*iter);
			index = text.lastIndexOf('/');
			
			if(!(index < 0))
			{
				text.truncate(index);
			}

			d_working_path.setPath(text);
			init = false;
			d_data_changed = true;
		}
	}
	
	d_action_export_pdf->setEnabled(d_data.Page_Count() > 0);
	d_action_export_images->setEnabled(d_data.Page_Count() > 0);
	d_action_reindex->setEnabled(d_data.Page_Count() > 0);
	d_action_save->setEnabled(d_data.Page_Count() > 0);
}

void TReportRoadMap::File_Save(void)
{
	if(d_active_file_name.length())
	{
		if(!this->Save_Data(d_active_file_name))
		{
			d_msg_box->setText("Error Saving Data");
			d_msg_box->setInformativeText("The Report Road Map file cannot be saved.");
			d_msg_box->setDetailedText(d_last_error);
			d_msg_box->setStandardButtons(QMessageBox::Ok);
			d_msg_box->setDefaultButton(QMessageBox::Ok);
			d_msg_box->setIcon(QMessageBox::Critical);
			d_msg_box->exec();
		}
		
		d_data_changed = false;

		return;
	}
	
	this->File_SaveAs();
}

void TReportRoadMap::File_SaveAs(void)
{
	QString								file_name;
	int									index;
	
	file_name = QFileDialog::getSaveFileName(this,
												  QStringLiteral("Save report roadmap file"),
												  d_working_path.absolutePath(),
												  QStringLiteral("Report Road Map Files (*.reportmap)"));
	
	if(file_name.length())
	{
		if(!this->Save_Data(file_name))
		{
			d_msg_box->setText("Error Saving Data");
			d_msg_box->setInformativeText("The Report Road Map file cannot be saved.");
			d_msg_box->setDetailedText(d_last_error);
			d_msg_box->setStandardButtons(QMessageBox::Ok);
			d_msg_box->setDefaultButton(QMessageBox::Ok);
			d_msg_box->setIcon(QMessageBox::Critical);
			d_msg_box->exec();
		}
		
		d_active_file_name = file_name;
		
		d_data_changed = false;
		
		index = file_name.lastIndexOf('/');
		
		if(!(index < 0))
		{
			file_name.truncate(index);
		}
		
		d_working_path.setPath(file_name);
	}
}

void TReportRoadMap::File_Export_Images(void)
{
	QString								file_path;
	QList<int>							page_ids;

	file_path = QFileDialog::getExistingDirectory(this,
											 QStringLiteral("Export image files"),
											 d_working_path.absolutePath(),
											 QFileDialog::ShowDirsOnly);
	
	if(file_path.length())
	{
		d_working_path.setPath(file_path);

		page_ids = d_page_widget->Page_Ids();

		this->Export_Images(file_path,page_ids);
	}
}

void TReportRoadMap::File_Export_Pdf(void)
{
	QList<int>							page_ids;
	QString								file_name;
	int									index;
	TReportSettings						settings;
	
	if(!(d_report_header_dialog))
	{
		d_report_header_dialog = new TReportHeaderDialog(this);
	}
	
	if(QDialog::Rejected == d_report_header_dialog->exec())
	{
		return;
	}

	
	file_name = QFileDialog::getSaveFileName(this,
											 QStringLiteral("Export PDF file"),
											 d_working_path.absolutePath(),
											 QStringLiteral("PDF File (*.pdf)"));
	
	if(file_name.length())
	{
		page_ids = d_page_widget->Page_Ids();
		
		settings.orientation = d_settings_widget->PDF_Orientation();
		settings.resolution = d_settings_widget->PDF_Resolution();
		settings.report_header1 = d_report_header_dialog->Header1_Text();
		settings.report_header2 = d_report_header_dialog->Header2_Text();

		this->Export_PDF(file_name,page_ids,settings);

		index = file_name.lastIndexOf('/');
		
		if(!(index < 0))
		{
			file_name.truncate(index);
		}
		
		d_working_path.setPath(file_name);
		
		if(settings.report_header1.length() || settings.report_header2.length())
		{
			d_data_changed = true;
		}
	}
}

void TReportRoadMap::Edit_Reindex(void)
{
	TDataItem							item;
	int									cntr;
	int									current_value;
	int									init_value;
	int									increment_over_page_method;
	int									reindex_option;
	int									active_page_id;
	QList<int>							page_ids;
	QList<int>::const_iterator			iter;
	
	if(!d_reindex_dialog)
	{
		d_reindex_dialog = new TReindexDialog(this);
	}
	
	if(QDialog::Accepted != d_reindex_dialog->exec())
	{
		return;
	}
	
	init_value = d_reindex_dialog->Initial_Value();
	increment_over_page_method = d_reindex_dialog->Increment_Over_Page_Method();
	reindex_option = d_reindex_dialog->Reindex_Option();
	
	current_value = init_value;
	
	if(reindex_option == TReindexOption::OPTION_ALL_ITEMS)
	{
		active_page_id = d_data.Active_Page();
		
		page_ids = d_page_widget->Page_Ids();
		
		for(iter = page_ids.begin();iter != page_ids.end();++iter)
		{
			d_data.Set_Active_Page(*iter);
			
			for(cntr = 0;cntr < d_data.Item_Count();++cntr)
			{
				item = d_data.Item(cntr);
				
				d_data.Change_Item_Value(item.id,current_value);
				current_value++;
			}
			
			// page increment
			switch(increment_over_page_method)
			{
				case TIncrementOverPageMethod::INCREMENT_1X:	// no change necessary
					break;
					
				case TIncrementOverPageMethod::INCREMENT_10X:
					current_value /= 10;
					current_value++;
					current_value *= 10;
					break;
					
				case TIncrementOverPageMethod::INCREMENT_100X:
					current_value /= 100;
					current_value++;
					current_value *= 100;
					break;
					
				case TIncrementOverPageMethod::INCREMENT_1000X:
					current_value /= 1000;
					current_value++;
					current_value *= 1000;
					break;
			}
		}
		
		d_data.Set_Active_Page(active_page_id);
		
		if(!(active_page_id < 0))
		{
			// update display
			
			d_display_widget->setUpdatesEnabled(false);

			d_label_widget->Clear();
			d_display_widget->Clear();
			
			d_display_widget->Set_Background_Image(QPixmap::fromImage(d_data.Page_Image()));
			
			for(cntr = 0;cntr < d_data.Item_Count();++cntr)
			{
				item = d_data.Item(cntr);
				
				d_label_widget->Add_Item(item);
				
				d_display_widget->Add_Item(item.id,item.value,item.text,item.border_type);
				d_display_widget->Move_Item(item.id,item.position);
			}
			
			d_display_widget->setUpdatesEnabled(true);
			
			d_label_widget->Set_Label_Index(current_value);
		}
	}
	else if(reindex_option == TReindexOption::OPTION_PAGE_ITEMS)
	{
		active_page_id = d_data.Active_Page();
		
		if(!(active_page_id < 0))
		{
			for(cntr = 0;cntr < d_data.Item_Count();++cntr)
			{
				item = d_data.Item(cntr);
				
				d_data.Change_Item_Value(item.id,current_value);
				current_value++;
			}
			
			// update display
			
			d_display_widget->setUpdatesEnabled(false);
			
			d_label_widget->Clear();
			d_display_widget->Clear();
			
			d_display_widget->Set_Background_Image(QPixmap::fromImage(d_data.Page_Image()));
			
			for(cntr = 0;cntr < d_data.Item_Count();++cntr)
			{
				item = d_data.Item(cntr);
				
				d_label_widget->Add_Item(item);
				
				d_display_widget->Add_Item(item.id,item.value,item.text,item.border_type);
				d_display_widget->Move_Item(item.id,item.position);
			}
			
			d_display_widget->setUpdatesEnabled(true);
			
			d_label_widget->Set_Label_Index(current_value);
		}
	}
	else if(reindex_option == TReindexOption::OPTION_SELECTED_ITEMS)
	{
		active_page_id = d_data.Active_Page();
		
		if(!(active_page_id < 0))
		{
			for(cntr = 0;cntr < d_data.Item_Count();++cntr)
			{
				item = d_data.Item(cntr);
				
				if(item.is_selected)
				{
					d_data.Change_Item_Value(item.id,current_value);
					current_value++;
				}
			}
			
			// update display
			
			d_display_widget->setUpdatesEnabled(false);
			
			d_label_widget->Clear();
			d_display_widget->Clear();
			
			d_display_widget->Set_Background_Image(QPixmap::fromImage(d_data.Page_Image()));
			
			for(cntr = 0;cntr < d_data.Item_Count();++cntr)
			{
				item = d_data.Item(cntr);
				
				d_label_widget->Add_Item(item);
				
				d_display_widget->Add_Item(item.id,item.value,item.text,item.border_type);
				d_display_widget->Move_Item(item.id,item.position);
			}
			
			d_display_widget->setUpdatesEnabled(true);
			
			d_label_widget->Set_Label_Index(current_value);
		}
	}
}

void TReportRoadMap::Help_About(void)
{
	d_msg_box->setText("Help - About");
	d_msg_box->setInformativeText("Report Roadmap - Version 3.0");
	d_msg_box->setDetailedText(QString());
	d_msg_box->setIcon(QMessageBox::Information);
	d_msg_box->setStandardButtons(QMessageBox::Ok);
	d_msg_box->setDefaultButton(QMessageBox::Ok);
	
	d_msg_box->exec();
}

void TReportRoadMap::Image_Zoom_Changed(
	int									value)
{
	double								scale_value;
	double								multiplier(0.13533528323661269189);	// A=1/exp(Bn) where B is 2.  See logarithmic volume controls
	
	// Convert the value to 1 +/- 1
	scale_value = static_cast<double>(value - d_zoom_slider->minimum()) / static_cast<double>(d_zoom_slider->maximum());
	
	scale_value = multiplier * exp(2 * scale_value);
	
	d_display_widget->Change_Scale(scale_value);
	
	d_zoom_scale_label->setText(QString("%1 X").arg(scale_value,0,'f',2));
}

void TReportRoadMap::Label_Size_Changed(
	int									value)
{
	int									font_size;
	
	font_size = static_cast<int>(0.5 + (static_cast<double>(value) * 0.4));
	
	d_display_widget->Set_Label_Size(QSize(value,value),font_size);
	
	d_data_changed = true;
}

void TReportRoadMap::Color_Changed(
	const QColor						color)
{
	d_display_widget->Set_Label_Color(color);
}

void TReportRoadMap::Item_Display_Format_Changed(
	const QString						&text)
{
	d_label_widget->Set_Display_Format(text);
	
	this->Page_Selection_Changed(d_data.Active_Page());
}

void TReportRoadMap::Track_Item_Option_Changed(
	bool								state)
{
	d_display_widget->Set_Track_Label(state);
}

void TReportRoadMap::Add_Item(void)
{
	int									id;
	int									label_index;
	int									border_type;
	TDataItem							data_item;
	int									count;
	static const int					MAX_COUNT(100);	// if this many random attempts failed then something is wrong
	
	if(d_data.Active_Page() < 0)
	{
		return;
	}
	
	if(!d_data.Page_Image_Valid())
	{
		return;
	}

	border_type = d_label_widget->Border_Type();
	label_index = d_label_widget->Next_Label_Index();
	
	// create a random label id.
	// display will return null if label already exists.
	
	srand(time(NULL));
	count = 0;

	do
	{
		id = rand();
		++count;
		
	}while(d_display_widget->Add_Item(id,label_index,QString(),border_type) == 0 && count < MAX_COUNT);
	
	if(count < MAX_COUNT)
	{	
		data_item.id = id;
		data_item.value = label_index;
		data_item.border_type = border_type;
		data_item.is_selected = false;
		data_item.position = d_display_widget->Item_Position(id);
		
		d_label_widget->Set_Label_Index(label_index + 1);
		
		d_data.Add_Item(data_item);
		d_label_widget->Add_Item(data_item);
	}
	
	d_data_changed = true;
}

void TReportRoadMap::Add_Item(
	const QPointF						&center_position)
{
	int									id;
	int									label_index;
	int									border_type;
	TDataItem							data_item;
	int									count;
	static const int					MAX_COUNT(100);	// if this many random attempts failed then something is wrong
	
	if(d_data.Active_Page() < 0)
	{
		return;
	}
	
	if(!d_data.Page_Image_Valid())
	{
		return;
	}

	border_type = d_label_widget->Border_Type();
	label_index = d_label_widget->Next_Label_Index();
	
	// create a random label id.
	// display will return null if label already exists.
	
	srand(time(NULL));
	count = 0;
	
	do
	{
		id = rand();
		++count;
		
	}while(d_display_widget->Add_Item(id,label_index,QString(),border_type) == 0 && count < MAX_COUNT);
	
	if(count < MAX_COUNT)
	{
		data_item.id = id;
		data_item.value = label_index;
		data_item.border_type = border_type;
		data_item.is_selected = false;
		
		d_label_widget->Set_Label_Index(label_index + 1);
	
		d_data.Add_Item(data_item);
		d_label_widget->Add_Item(data_item);
		
		d_display_widget->Move_Item(id,center_position);
	}
	
	d_data_changed = true;
}

void TReportRoadMap::Remove_Item(void)
{
	std::vector<int>					list;
	std::vector<int>::const_iterator	iter;
	
	list = d_label_widget->Selected_Items();
	
	if(!list.size())
	{
		return;
	}
	
	d_msg_box->setText("Remove Labels");
	d_msg_box->setInformativeText("Labels will be deleted.  Are you sure?");
	d_msg_box->setDetailedText(QString());
	d_msg_box->setStandardButtons(QMessageBox::Yes | QMessageBox::No);
	d_msg_box->setDefaultButton(QMessageBox::No);
	d_msg_box->setIcon(QMessageBox::Question);
	
	if(d_msg_box->exec() == QMessageBox::No)
	{
		return;
	}
	
	for(iter = list.begin();iter != list.end();++iter)
	{
		d_display_widget->Remove_Item(*iter);
		d_data.Remove_Item(*iter);
		d_label_widget->Remove_Item(*iter);
	}
	
	d_data_changed = true;
}

void TReportRoadMap::Change_Item_Value(
	const int							id,
	const int							new_value)
{
	d_data.Change_Item_Value(id,new_value);
	d_display_widget->Change_Item_Value(id,new_value);
	
	d_data_changed = true;
}

void TReportRoadMap::Change_Item_Text(
	const int							id,
	const QString						&new_text)
{
	d_data.Change_Item_Text(id,new_text);
	d_display_widget->Change_Item_Text(id,new_text);
	
	d_data_changed = true;
}

void TReportRoadMap::Change_Item_Border(
	const int							id,
	const int							new_border_type)
{
	d_data.Change_Item_Border(id,new_border_type);
	d_display_widget->Change_Item_Border(id,new_border_type);
	
	d_data_changed = true;
}

void TReportRoadMap::Item_Selection_Changed(void)
{
	std::vector<int>					sel_ids;
	std::vector<int>::const_iterator	iter;

	sel_ids = d_label_widget->Selected_Items();
	
	d_display_widget->Set_Selected_Items(sel_ids,!d_block_label_tracking);
	
	if(!(d_data.Active_Page() < 0))
	{
		d_data.Clear_Selection();
		for(iter = sel_ids.begin();iter != sel_ids.end();++iter)
		{
			d_data.Change_Item_Selection((*iter),true);
		}
	}
}

void TReportRoadMap::Display_Item_Clicked(
	const int							id)
{
	d_block_label_tracking = true;
	d_label_widget->Set_Selection(id);
	d_block_label_tracking = false;
	
	if(!(d_data.Active_Page() < 0))
	{
		d_data.Clear_Selection();
		d_data.Change_Item_Selection(id,true);
	}
}

void TReportRoadMap::Update_Item_Position(
	const int							id,
	const QPointF						&position)
{
	if(!(d_data.Active_Page() < 0))
	{
		d_data.Change_Item_Position(id,position);
		d_data_changed = true;
	}
}

void TReportRoadMap::Add_Page(void)
{
	int									page_id;
	
	page_id = d_data.Add_Page();
	d_page_widget->Add_Page(page_id);
	
	d_data_changed = true;
}

void TReportRoadMap::Remove_Pages(
	const QList<int>					&page_ids)
{
	QList<int>::const_iterator			iter;
	
	d_msg_box->setText("Remove Pages");
	d_msg_box->setInformativeText("All page associated labels will be deleted.  Are you sure?");
	d_msg_box->setDetailedText(QString());
	d_msg_box->setStandardButtons(QMessageBox::Yes | QMessageBox::No);
	d_msg_box->setDefaultButton(QMessageBox::No);
	d_msg_box->setIcon(QMessageBox::Question);
	
	if(d_msg_box->exec() == QMessageBox::Yes)
	{
		d_action_export_images->setEnabled(false);
		d_action_export_pdf->setEnabled(false);
		d_action_reindex->setEnabled(false);
		d_action_save->setEnabled(false);
		
		d_label_widget->Clear();
		
		d_display_widget->Clear();
		d_display_widget->Set_Background_Image(QPixmap(":/images/background_default.png"));
		
		for(iter = page_ids.begin();iter != page_ids.end();++iter)
		{
			d_data.Remove_Page(*iter);
			d_page_widget->Remove_Page(*iter);
		}
	}
	
	d_action_export_pdf->setEnabled(d_data.Page_Count() > 0);
	d_action_export_images->setEnabled(d_data.Page_Count() > 0);
	d_action_reindex->setEnabled(d_data.Page_Count() > 0);
	d_action_save->setEnabled(d_data.Page_Count() > 0);

	d_data_changed = true;
}

void TReportRoadMap::Page_Selection_Changed(
	int									page_id)
{
	TDataItem							data_item;
	int									count;
	int									cntr;
	bool								prev_state;
	
	d_label_widget->Clear();
	d_display_widget->Clear();
	
	prev_state = d_data_changed;		// push_current state
	
	if(d_data.Set_Active_Page(page_id))
	{
		if(d_data.Page_Image_Valid())
		{
			d_label_widget->setEnabled(true);
			
			d_display_widget->setUpdatesEnabled(false);
			
			d_display_widget->Set_Background_Image(QPixmap::fromImage(d_data.Page_Image()));
			
			count = d_data.Item_Count();
			
			for(cntr = 0;cntr < count;++cntr)
			{
				data_item = d_data.Item(cntr);
				
				d_label_widget->Add_Item(data_item);
				d_display_widget->Add_Item(data_item.id,data_item.value,data_item.text,data_item.border_type);
				d_display_widget->Move_Item(data_item.id,data_item.position);
			}
			
			d_display_widget->setUpdatesEnabled(true);
		}
		else
		{
			d_display_widget->Set_Background_Image(QPixmap(":/images/background_default.png"));
			d_label_widget->setEnabled(false);
		}
	}
	else
	{
		d_display_widget->Set_Background_Image(QPixmap(":/images/background_default.png"));
		d_label_widget->setEnabled(false);
	}
	
	d_data_changed = prev_state;		// push_current state
}

void TReportRoadMap::closeEvent(
	QCloseEvent							*event)
{
	d_settings->setValue("Mainwindow_Position", this->pos());
	d_settings->setValue("Mainwindow_Size", this->size());

	d_settings->setValue("Working_Path", d_working_path.absolutePath());

	d_settings->setValue("Track_Selected_Label", d_settings_widget->Track_Label());
	d_settings->setValue("Label_Size", d_settings_widget->Label_Size());
	d_settings->setValue("Item_Display_Format", d_settings_widget->Display_Format());
	d_settings->setValue("Label_Color", d_settings_widget->Label_Color().name());
	d_settings->setValue("PDF_Orientation", d_settings_widget->PDF_Orientation());
	d_settings->setValue("PDF_Resolution", d_settings_widget->PDF_Resolution());
	
	d_settings->setValue("Border_Type", d_label_widget->Border_Type());
	
	d_settings->setValue("Label_Visible", d_label_dock->isVisible());
	d_settings->setValue("Settings_Visible", d_settings_dock->isVisible());
	d_settings->setValue("Page_Visible", d_page_dock->isVisible());
	
	d_settings->setValue("Label_Floating", d_label_dock->isFloating());
	d_settings->setValue("Settings_Floating", d_settings_dock->isFloating());
	d_settings->setValue("Page_Floating", d_page_dock->isFloating());
	
	d_settings->setValue("Label_Position", d_label_dock->pos());
	d_settings->setValue("Settings_Position", d_settings_dock->pos());
	d_settings->setValue("Page_Position", d_page_dock->pos());

	if(d_data_changed)
	{
		d_msg_box->setText("Data Not Saved");
		d_msg_box->setInformativeText("Do you want to save the Report RoadMap data?");
		d_msg_box->setDetailedText(QString());
		d_msg_box->setStandardButtons(QMessageBox::Yes | QMessageBox::No | QMessageBox::Cancel);
		d_msg_box->setDefaultButton(QMessageBox::Yes);
		d_msg_box->setIcon(QMessageBox::Question);
		
		switch(d_msg_box->exec())
		{
			case QMessageBox::Yes:
				this->File_Save();
				break;
				
			case QMessageBox::Cancel:
				event->ignore();
				return;
				
			default:
				break;
		}

	}
	
	event->accept();
}

void TReportRoadMap::Load_Image(
	const QString						&file_name)
{
	QImage								image;
	int									active_page_id;
	int									page_id;
	
	if(image.load(file_name))
	{
		active_page_id = d_data.Active_Page();
		
		if(active_page_id < 0)
		{
			page_id = d_data.Add_Page(image);
			d_page_widget->Add_Page(page_id);
		}
		else
		{

			if(d_data.Page_Has_Data())
			{
				page_id = d_data.Add_Page(image);
				d_page_widget->Add_Page(page_id);
			}
			else
			{
				d_data.Set_Page_Image(image);
				d_display_widget->Set_Background_Image(QPixmap::fromImage(image));
				d_label_widget->setEnabled(true);
			}
		}
	}
}


bool TReportRoadMap::Load_Legacy_Data(
	const QString 						&file_name)
{
	TXmlFile							xml_file;
	int									version_number;
	TDataItem							item;
	std::vector<TDataItem>				items;
	std::vector<TDataItem>::const_iterator iter;
	QImage								image;
	int									page_id;
	int									max_value(0);
	
	if(!image.load(QString("%1.rmi").arg(file_name),"PNG"))
	{
		d_last_error = QStringLiteral("Background image not found");
		return false;
	}
	
	try
	{
		xml_file.Set_File_Name(QString("%1.rmd").arg(file_name));
		xml_file.Open(QStringLiteral("Report_RoadMap_Data"),&version_number);
		
		switch(version_number)
		{
			case 1:
				this->Load_Legacy_V1(&xml_file,&items);
				break;
				
			case 2:
				this->Load_Legacy_V2(&xml_file,&items);
				break;
		}
		
	}
	catch(TXmlFileException XmlFileException)
	{
		d_last_error = XmlFileException.ErrorText();
		return false;
	}
	
	d_data.Clear();
	d_page_widget->Clear();
	
	page_id = d_data.Add_Page(image);
	
	if(!(page_id < 0))
	{
		d_page_widget->Add_Page(page_id);
		
		d_data.Set_Active_Page(page_id);
		
		for(iter = items.begin();iter != items.end();++iter)
		{
			d_data.Add_Item((*iter));
			
			if((*iter).value > max_value)
			{
				max_value = (*iter).value;
			}
		}
		
		d_page_widget->Set_Active_Page(page_id);
		d_label_widget->Set_Label_Index(max_value + 1);
	}

	return true;
}

void TReportRoadMap::Load_Legacy_V1(
	TXmlFile 							*xml_file,
	std::vector<TDataItem>				*items)
{
	TDataItem							item;
	QDomElement							el0;
	QDomElement							el1;
	QString								text;
	int									width;
	int									ival;
	QStringList							pos_items;
	QColor								color;
	
	assert(xml_file);
	
	// Item Size
	el0 = xml_file->Get_Node(xml_file->Root_Node(),QStringLiteral("Item_Size"));
	
	el1 = xml_file->Get_Node(&el0,QStringLiteral("Width"));
	width = xml_file->Read_Text_Node(&el1).toInt();
	
	if(width < 16 || width > 128)
	{
		width = 32;
	}
	
	d_settings_widget->Set_Label_Size(width);
	
	// Color
	el0 = xml_file->Get_Node(xml_file->Root_Node(),QStringLiteral("Colors"));
	
	el1 = xml_file->Get_Node(&el0,QStringLiteral("Foreground"));
	color.setNamedColor(xml_file->Read_Text_Node(&el1));
	
	d_settings_widget->Set_Label_Color(color);
	d_display_widget->Set_Label_Color(color);

	el0 = xml_file->Get_Node(xml_file->Root_Node(),"Item",true);
	
	while(!el0.isNull())
	{
		// item id
		el1 = xml_file->Get_Node(&el0,QStringLiteral("ID"));
		item.id = xml_file->Read_Text_Node(&el1).toInt();
		
		// Version 1 doesn't have id_text field
		item.id = QString("%1").arg(item.id).toInt();
		item.value = item.id;
		
		el1 = xml_file->Get_Node(&el0,QStringLiteral("Border_Type"));
		ival = static_cast<int>(xml_file->Read_Text_Node(&el1).toInt());
		
		switch(ival)
		{
			case 0:
			default:
				item.border_type = TBorderType::BORDER_TYPE_NONE;
				break;
				
			case 1:
				item.border_type = TBorderType::BORDER_TYPE_SQUARE;
				break;
				
			case 2:
				item.border_type = TBorderType::BORDER_TYPE_ELLIPSE;
				break;
				
			case 3:
				item.border_type = TBorderType::BORDER_TYPE_TRIANGLE;
				break;
		}
		
		el1 = xml_file->Get_Node(&el0,QStringLiteral("Position"));
		text = xml_file->Read_Text_Node(&el1);
		
		pos_items = text.split(',');
		
		if(pos_items.size() == 2)
		{
			item.position = QPointF(pos_items[0].toDouble(),pos_items[1].toDouble());
			items->push_back(item);
		}
		
		el0 = xml_file->Get_Sibling_Node(&el0,QStringLiteral("Item"),true);
	}
}

void TReportRoadMap::Load_Legacy_V2(
	TXmlFile 							*xml_file,
	std::vector<TDataItem>				*items)
{
	TDataItem							item;
	QDomElement							el0;
	QDomElement							el1;
	QString								text;
	int									width;
	int									ival;
	bool								state;
	QStringList							pos_items;
	QColor								color;

	assert(xml_file);
	
	// Item Size
	el0 = xml_file->Get_Node(xml_file->Root_Node(),QStringLiteral("Item_Size"));
	
	el1 = xml_file->Get_Node(&el0,QStringLiteral("Width"));
	width = xml_file->Read_Text_Node(&el1).toInt();
	
	if(width < 16 || width > 128)
	{
		width = 32;
	}
	d_settings_widget->Set_Label_Size(width);
	
	// Color
	el0 = xml_file->Get_Node(xml_file->Root_Node(),QStringLiteral("Colors"));
	
	el1 = xml_file->Get_Node(&el0,QStringLiteral("Foreground"));
	color.setNamedColor(xml_file->Read_Text_Node(&el1));
	
	d_settings_widget->Set_Label_Color(color);
	d_display_widget->Set_Label_Color(color);

	el0 = xml_file->Get_Node(xml_file->Root_Node(),QStringLiteral("Item"),true);
	
	while(!el0.isNull())
	{
		// item id
		el1 = xml_file->Get_Node(&el0,"ID");
		item.id = xml_file->Read_Text_Node(&el1).toInt();
		item.value = item.id;
		
		el1 = xml_file->Get_Node(&el0,QStringLiteral("ID_Text"));
		text = xml_file->Read_Text_Node(&el1);
	
		// check if node contains a numerical value or actual text.
		ival = text.toInt(&state);
		
		if(!state)
		{
			item.text = xml_file->Read_Text_Node(&el1);
		}
		else
		{
			item.text = QString();
		}
		
		el1 = xml_file->Get_Node(&el0,QStringLiteral("Border_Type"));
		ival = static_cast<int>(xml_file->Read_Text_Node(&el1).toInt());
		
		switch(ival)
		{
			case 0:
			default:
				item.border_type = TBorderType::BORDER_TYPE_NONE;
				break;
				
			case 1:
				item.border_type = TBorderType::BORDER_TYPE_SQUARE;
				break;
				
			case 2:
				item.border_type = TBorderType::BORDER_TYPE_ELLIPSE;
				break;
				
			case 3:
				item.border_type = TBorderType::BORDER_TYPE_TRIANGLE;
				break;
		}

		el1 = xml_file->Get_Node(&el0,QStringLiteral("Position"));
		text = xml_file->Read_Text_Node(&el1);
		
		pos_items = text.split(',');
		
		if(pos_items.size() == 2)
		{
			item.position = QPointF(pos_items[0].toDouble(),pos_items[1].toDouble());
			items->push_back(item);
		}
		
		el0 = xml_file->Get_Sibling_Node(&el0,QStringLiteral("Item"),true);
	}
}

bool TReportRoadMap::Save_Data(
	const QString						&file_name)
{
	QFile								file;
	QByteArray							file_data;
	QByteArray							data;
	QImage								image;
	QString								text;
	QString								header1;
	QString								header2;
	QList<int>							page_ids;
	QList<int>::const_iterator			iter;
	int									active_page_id;
	int									ival;
	unsigned char						*pival(reinterpret_cast<unsigned char*>(&ival));
	
	try
	{
		file.setFileName(file_name);
		if(!file.open(QIODevice::WriteOnly))
		{
			d_last_error = QStringLiteral("Cannot open file for writing");
			return false;
		}
		
		// magic number
		ival = 90210;
		file_data.append(reinterpret_cast<char*>(pival),sizeof(int));
		
		// item label size
		ival = d_settings_widget->Label_Size();
		file_data.append(reinterpret_cast<char*>(pival),sizeof(int));
		
		// item label color, size then text
		text = d_settings_widget->Label_Color().name();
		
		ival = text.size();
		file_data.append(reinterpret_cast<char*>(pival),sizeof(int));
		file_data.append(text.toLatin1());
		
		// report headers, size then text
		if(d_report_header_dialog)
		{
			header1 = d_report_header_dialog->Header1_Text();
			header2 = d_report_header_dialog->Header2_Text();
		}
		
		ival = header1.size();
		file_data.append(reinterpret_cast<char*>(pival),sizeof(int));
		file_data.append(header1.toLatin1());
		
		ival = header2.size();
		file_data.append(reinterpret_cast<char*>(pival),sizeof(int));
		file_data.append(header2.toLatin1());

		// write number of pages
		ival = d_data.Page_Count();
		file_data.append(reinterpret_cast<char*>(pival),sizeof(int));

		active_page_id = d_data.Active_Page();
		
		page_ids = d_page_widget->Page_Ids();

		for(iter = page_ids.begin();iter != page_ids.end();++iter)
		{
			// page name, length then contents
			text = d_page_widget->Page_Name(*iter);
						
			ival = text.size();
			file_data.append(reinterpret_cast<char*>(pival),sizeof(int));
			file_data.append(text.toLatin1());
			
			d_data.Set_Active_Page(*iter);
			
			image = d_data.Page_Image();
			
			QBuffer buffer(&data);
			
			buffer.open(QIODevice::WriteOnly);
			image.save(&buffer, "PNG");
			
			buffer.close();
			
			// size of image
			ival = data.size();
			file_data.append(reinterpret_cast<char*>(pival),sizeof(int));

			// image_data
			file_data.append(data);
			
			data = this->Write_Items();
			
			// size of xml
			ival = data.size();
			file_data.append(reinterpret_cast<char*>(pival),sizeof(int));
			
			// xml
			file_data.append(data);
		}
		
		d_data.Set_Active_Page(active_page_id);

		// write file
		file.write(file_data);
		file.close();
	}
	catch(TXmlFileException exception)
	{
		d_last_error = exception.ErrorText();
		return false;
	}
	
	return true;
}

QByteArray TReportRoadMap::Write_Items(void)
{
	TXmlFile							xmlfile;
	QDomElement							e0;
	QDomElement							e1;
	QByteArray							data;
	TDataItem							item;
	int									cntr;
	
	xmlfile.Reset("Report_RoadMap_Data",1);

	for(cntr = 0;cntr < d_data.Item_Count();++cntr)
	{
		item = d_data.Item(cntr);
		
		e0 = xmlfile.Create_Node(xmlfile.Root_Node(),QStringLiteral("Item"));
		
		// id
		e1 = xmlfile.Create_Node(&e0,QStringLiteral("Id"));
		xmlfile.Write_Text_Node(&e1,item.id);
		
		// value
		e1 = xmlfile.Create_Node(&e0,QStringLiteral("Value"));
		xmlfile.Write_Text_Node(&e1,item.value);
		
		// text
		e1 = xmlfile.Create_Node(&e0,QStringLiteral("Text"));
		xmlfile.Write_Text_Node(&e1,item.text);
		
		// border_type
		e1 = xmlfile.Create_Node(&e0,QStringLiteral("Border_Type"));
		xmlfile.Write_Text_Node(&e1,item.border_type);
		
		// position
		e1 = xmlfile.Create_Node(&e0,QStringLiteral("Position"));
		xmlfile.Write_Text_Node(&e1,QString("%1,%2")
							  .arg(item.position.x(),0,'f',2)
							  .arg(item.position.y(),0,'f',2));

	}
	
	xmlfile.Write_Data(&data,true);
	
	return data;
}

bool TReportRoadMap::Load_Data(
	const QString						&file_name)
{
	QFile								file;
	QByteArray							file_data;
	QByteArray							data;
	QImage								image;
	QString								text;
	QString								header1;
	QString								header2;
	QString								page_name;
	QColor								color;
	int									ival;
	int									index;
	int									count;
	int									cntr;
	int									page_id(-1);
	int									max_value(0);

	file.setFileName(file_name);
	if(!file.open(QIODevice::ReadOnly))
	{
		d_last_error = QStringLiteral("Cannot open file for reading");
		return false;
	}
	
	file_data = file.readAll();
	file.close();
	
	if(file_data.size() < 8)
	{
		d_last_error = QStringLiteral("Invalid file.");
		return false;
	}
	
	// magic number
	index = 0;
	ival = Conversion::Integer(index,file_data);
	index += sizeof(int);
	
	if(ival != 90210)
	{
		d_last_error = QStringLiteral("Invalid file.");
		return false;
	}
	
	// label_size
	ival = Conversion::Integer(index,file_data);
	index += sizeof(int);
	
	d_settings_widget->Set_Label_Size(ival);
	
	// label color, size first then text
	
	ival = Conversion::Integer(index,file_data);
	index += sizeof(int);
	
	text = QString::fromLatin1(file_data.mid(index,ival));
	index += ival;

	color.setNamedColor(text);
	
	d_settings_widget->Set_Label_Color(color);
	d_display_widget->Set_Label_Color(color);
	

	// headers, size first then text
	ival = Conversion::Integer(index,file_data);
	index += sizeof(int);
	
	header1 = QString::fromLatin1(file_data.mid(index,ival));
	index += ival;
	
	ival = Conversion::Integer(index,file_data);
	index += sizeof(int);
	
	header2 = QString::fromLatin1(file_data.mid(index,ival));
	index += ival;
	
	if(header1.length() || header2.length())
	{
		if(!d_report_header_dialog)
		{
			d_report_header_dialog = new TReportHeaderDialog(this);
		}
		
		d_report_header_dialog->Set_Header1_Text(header1);
		d_report_header_dialog->Set_Header2_Text(header2);
	}

	d_data.Clear();
	d_page_widget->Clear();
	
	// page count
	count = Conversion::Integer(index,file_data);
	index += sizeof(int);
	
	for(cntr = 0;cntr < count;++cntr)
	{
		// page name		
		ival = Conversion::Integer(index,file_data);
		index += sizeof(int);
		
		page_name = QString::fromLatin1(file_data.mid(index,ival));
		index += ival;

		// image size
		ival = Conversion::Integer(index,file_data);
		index += sizeof(int);
		
		data = file_data.mid(index,ival);
		index += ival;
		
		image.loadFromData(data);
		
		page_id = d_data.Add_Page(image);
		
		if(page_id < 0)
		{
			d_last_error = QStringLiteral("Invalid file.");
			return false;
		}
		
		d_data.Set_Active_Page(page_id);
		
		// data size
		ival = Conversion::Integer(index,file_data);
		index += sizeof(int);
		
		data = file_data.mid(index,ival);
		index += ival;
		
		d_page_widget->Add_Page(page_id);
		d_page_widget->Set_Page_Name(page_id,page_name);

		try
		{
			this->Read_Items(data);
			
			ival = d_data.Item_Max_Value();
			
			if(ival > max_value)
			{
				max_value = ival;
			}
		}
		catch(TXmlFileException xml_exception)
		{
			d_last_error = xml_exception.ErrorText();
			return false;
		}
	}
	
	d_page_widget->Set_Active_Page(page_id);
	d_label_widget->Set_Label_Index(max_value + 1);

	return true;
}

void TReportRoadMap::Read_Items(
	const QByteArray					&data)
{
	TXmlFile							xmlfile;
	QDomElement							e0;
	QDomElement							e1;
	TDataItem							item;
	QString								text;
	QStringList							list;
	int									version;

	xmlfile.Open_Data(data,QStringLiteral("Report_RoadMap_Data"),&version);
	
	if(version != 1)
	{
		throw TXmlFileException(QStringLiteral("XML Version not recognized"));
	}
	
	item.is_selected = false;
	
	e0 = xmlfile.Get_Node(xmlfile.Root_Node(),QStringLiteral("Item"),true);
	
	while(!e0.isNull())
	{
		e1 = xmlfile.Get_Node(&e0,QStringLiteral("Id"));
		item.id = xmlfile.Read_Text_Node(&e1).toInt();
		
		e1 = xmlfile.Get_Node(&e0,QStringLiteral("Value"));
		item.value = xmlfile.Read_Text_Node(&e1).toInt();
		
		e1 = xmlfile.Get_Node(&e0,QStringLiteral("Text"));
		item.text = xmlfile.Read_Text_Node(&e1,true);
		
		e1 = xmlfile.Get_Node(&e0,QStringLiteral("Border_Type"));
		item.border_type = xmlfile.Read_Text_Node(&e1).toInt();
		
		e1 = xmlfile.Get_Node(&e0,QStringLiteral("Position"));
		text = xmlfile.Read_Text_Node(&e1);
		
		list = text.split(',');
		
		if(list.size() != 2)
		{
			throw TXmlFileException(QStringLiteral("XML Item position invalid"));
		}
		
		item.position.setX(list[0].toDouble());
		item.position.setY(list[1].toDouble());
		
		d_data.Add_Item(item);
		
		e0 = xmlfile.Get_Sibling_Node(&e0,QStringLiteral("Item"),true);
	}
}

bool TReportRoadMap::Export_PDF(
	const QString						&file_name,
	const QList<int>					&page_ids,
	const TReportSettings				&settings)
{
	static const int					COLUMN_COUNT(4);
	static const int					ROW_SECTION_COUNT(4);	// 2 header lines, 1 image, 1 footer line
	QPainter							painter;
	QPrinter							printer(QPrinter::HighResolution);
	QPixmap								pixmap;
	QPixmap								pixmap_resized;
	QRect								page_rect;
	QRect								rect;
	QRect								pixmap_rect;
	QSize								pixmap_size;
	QSize								size;
	QFont								title_font("Helvetica",12,QFont::Normal);
	QFont								footer_font("Helvetica",6,QFont::Normal);
	QPen								border_pen(QColor(0,0,0));
	QPen								black_pen(QColor(0,0,0));
	QString								footer_text;
	TDataItem							data_item;
	QList<int>::const_iterator			iter;
	bool								init_page;
	int									ival;
	int									current_active_page;
	int									cntr;
	int									column_width;
	int									page_margin_width;
	int									page_margin_height;
	int									page_number;
	int									column_positions[COLUMN_COUNT];
	int									row_positions[ROW_SECTION_COUNT];
	
	if(page_ids.size() == 0)
	{
		return false;
	}
	
	if(settings.resolution < 1 || settings.resolution > 4)
	{
		return false;
	}
	
	if(settings.orientation)
	{
		printer.setOrientation(QPrinter::Landscape);
	}
	else
	{
		printer.setOrientation(QPrinter::Portrait);
	}
	
	printer.setOutputFormat(QPrinter::PdfFormat);
	printer.setOutputFileName(file_name);
	printer.setPageMargins(0.0,0.0,0.0,0.0,QPrinter::Millimeter);
	//	printer.setColorMode(QPrinter::GrayScale);
	printer.setResolution(96 * settings.resolution);		// higher dpi increases resolution.
	
	if(!painter.begin(&printer))
	{
		return false;
	}
	
	switch(settings.resolution)
	{
		case 1:
		default:
			border_pen.setWidth(1);
			break;
			
		case 2:
			border_pen.setWidth(2);
			break;
			
		case 3:
			border_pen.setWidth(2);
			break;
			
		case 4:
			border_pen.setWidth(3);
			break;
	}
	
	
	footer_text = QStringLiteral("Select Calibration Incorporated.  www.selectcalibration.ca");
	
	painter.setPen(black_pen);
	
	page_rect = printer.pageRect();
	
	page_margin_width = static_cast<int>(static_cast<double>(page_rect.width()) * 0.09);	// estimation based on 8.5x11 with margin of 3/4" for left and right
	page_margin_height = static_cast<int>(static_cast<double>(page_rect.height()) * 0.045);	// estimation based on 8.5x11 with margin of 1/2" for top and bottom
	
	page_rect.adjust(page_margin_width,page_margin_height,-page_margin_width,-page_margin_height);
	
	// calculate column positions
	
	column_width = static_cast<int>((static_cast<double>(page_rect.width()) / (COLUMN_COUNT - 1)) + 0.5);
	
	for(cntr = 0;cntr < COLUMN_COUNT;++cntr)
	{
		if(cntr == 0)
		{
			column_positions[0] = page_rect.left();
		}
		else
		{
			column_positions[cntr] = column_positions[cntr - 1] + column_width;
		}
	}
	
	// calculate row positions
	
	painter.setFont(title_font);
	QFontMetrics title_font_metrics = painter.fontMetrics();

	row_positions[0] = page_rect.top();
	
	if(settings.report_header1.length())
	{
		row_positions[1] = row_positions[0] + title_font_metrics.height();
		
		if(!settings.report_header2.length())
		{
			row_positions[1] += title_font_metrics.xHeight();
		}
	}
	else
	{
		row_positions[1] = row_positions[0];
	}
	
	if(settings.report_header2.length())
	{
		row_positions[2] = row_positions[1] + title_font_metrics.height() + title_font_metrics.xHeight();
	}
	else
	{
		row_positions[2] = row_positions[1];
	}

	
	painter.setFont(footer_font);
	QFontMetrics footer_font_metrics = painter.fontMetrics();
	
	row_positions[3] = page_rect.bottom() - footer_font_metrics.height() - footer_font_metrics.xHeight();
	
	pixmap_rect.setRect(column_positions[0],row_positions[2],column_positions[COLUMN_COUNT-1]-column_positions[0],row_positions[3] - row_positions[2]);

	init_page = false;
	
	current_active_page = d_data.Active_Page();
	
	d_display_widget->setUpdatesEnabled(false);

	for(iter = page_ids.begin();iter != page_ids.end();++iter)
	{
		if(init_page)
		{
			printer.newPage();
			page_number++;
			
			page_rect = printer.pageRect();
			
			page_margin_width = static_cast<int>(static_cast<double>(page_rect.width()) * 0.09);	// estimation based on 8.5x11 with margin of 3/4" for left and right
			page_margin_height = static_cast<int>(static_cast<double>(page_rect.height()) * 0.045);	// estimation based on 8.5x11 with margin of 1/2" for top and bottom
			
			page_rect.adjust(page_margin_width,page_margin_height,-page_margin_width,-page_margin_height);
		}
		else
		{
			init_page = true;
			
			page_number = 1;
		}
		
		
		d_data.Set_Active_Page((*iter));

		// draw upper and lower separator lines
		painter.setPen(border_pen);
		
		painter.drawLine(pixmap_rect.bottomLeft(),pixmap_rect.bottomRight());
		painter.drawLine(pixmap_rect.topLeft(),pixmap_rect.topRight());
		
		painter.setPen(black_pen);
		

		// header and footer
		painter.setFont(title_font);
		
		if(settings.report_header1.length())
		{
			rect.setRect(column_positions[0],row_positions[0],page_rect.width(),row_positions[1] - row_positions[0]);
			painter.drawText(rect,Qt::AlignLeft | Qt::AlignTop,settings.report_header1);
		}
	
		if(settings.report_header2.length())
		{
			rect.setRect(column_positions[0],row_positions[1],page_rect.width(),row_positions[2] - row_positions[1]);
			painter.drawText(rect,Qt::AlignLeft | Qt::AlignTop,settings.report_header2);
		}
		
		painter.setFont(footer_font);
		
		rect.setRect(column_positions[0],row_positions[3],page_rect.width(),page_rect.bottom() - row_positions[3]);

		painter.drawText(rect,Qt::AlignLeft | Qt::AlignBottom,footer_text);
		painter.drawText(rect,Qt::AlignRight | Qt::AlignBottom,QString("Page %1").arg(page_number));

		
		d_display_widget->Clear();
		d_display_widget->Set_Background_Image(QPixmap::fromImage(d_data.Page_Image()));
		
		for(cntr = 0;cntr < d_data.Item_Count();++cntr)
		{
			data_item = d_data.Item(cntr);
			
			d_display_widget->Add_Item(data_item.id,data_item.value,data_item.text,data_item.border_type);
			d_display_widget->Move_Item(data_item.id,data_item.position);
		}
		
		pixmap = d_display_widget->Pixmap();
		
		rect = pixmap_rect;
		rect.adjust(settings.resolution,settings.resolution,-settings.resolution,-settings.resolution);			// same value as display precision multiplier

		pixmap_resized = pixmap.scaled(rect.size(),Qt::KeepAspectRatio,Qt::SmoothTransformation);
		
		pixmap_size = pixmap_resized.size();
		size = rect.size();
		
		ival = size.width() - pixmap_size.width();
		if(ival > 2)
		{
			rect.adjust(ival/2,0,-ival/2,0);
		}
		
		ival = size.height() - pixmap_size.height();
		if(ival > 2)
		{
			rect.adjust(0,ival/2,0,-ival/2);
		}
		
		painter.drawPixmap(rect,pixmap_resized);
	}
	
	// restore original view in display widget
	d_data.Set_Active_Page(current_active_page);
	
	d_display_widget->Clear();
	d_display_widget->Set_Background_Image(QPixmap::fromImage(d_data.Page_Image()));
	
	for(cntr = 0;cntr < d_data.Item_Count();++cntr)
	{
		data_item = d_data.Item(cntr);
		
		d_display_widget->Add_Item(data_item.id,data_item.value,data_item.text,data_item.border_type);
		d_display_widget->Move_Item(data_item.id,data_item.position);
	}

	d_display_widget->setUpdatesEnabled(true);
	
	painter.end();
	
	return true;
}

bool TReportRoadMap::Export_Images(
	const QString						&file_path,
	const QList<int>					&page_ids)
{
	QList<int>::const_iterator			iter;
	QString								file_name;
	QString								text;
	QPixmap								pixmap;
	TDataItem							data_item;
	int									cntr;
	int									page_cntr;
	int									current_active_page;
	bool								return_state;
	QDateTime							date_time(QDateTime::currentDateTime());

	
	if(page_ids.size() == 0)
	{
		return false;
	}

	current_active_page = d_data.Active_Page();
	
	d_display_widget->setUpdatesEnabled(false);
	
	return_state = true;
	page_cntr = 1;
	
	for(iter = page_ids.begin();iter != page_ids.end();++iter)
	{
		d_data.Set_Active_Page((*iter));

		d_display_widget->Clear();
		d_display_widget->Set_Background_Image(QPixmap::fromImage(d_data.Page_Image()));
		
		for(cntr = 0;cntr < d_data.Item_Count();++cntr)
		{
			data_item = d_data.Item(cntr);
			
			d_display_widget->Add_Item(data_item.id,data_item.value,data_item.text,data_item.border_type);
			d_display_widget->Move_Item(data_item.id,data_item.position);
		}
		
		pixmap = d_display_widget->Pixmap();
		file_name = file_path;
		
		if(!(file_path.endsWith('/') || file_path.endsWith('\\')))
		{
			file_name.append('/');
		}
		
		text = d_page_widget->Page_Name(*iter);
		
		if(text.length() == 0)
		{
			// use generic name
			text = QString("Image_%1").arg(*iter);
		}
		else
		{
			// remove all char's that shouldn't be in a file name
			text.replace(QChar('/'),QChar('_'));
			text.replace(QChar('\\'),QChar('_'));
			text.replace(QChar(':'),QChar('_'));
			text.replace(QChar('>'),QChar('_'));
			text.replace(QChar('<'),QChar('_'));
		}
		
		// add date and time to file name
		text += QString("-%1").arg(date_time.toString("ddMMyyyyhhmmss"));
		
		// add page number and extension to file name
		text += QString("%1.png").arg(page_cntr);
		
		file_name += text;
		
		if(!pixmap.save(file_name,"png"))
		{
			return_state = false;
			break;
		}
		
		++page_cntr;
	}
	
	// restore original view
	d_data.Set_Active_Page(current_active_page);
	
	d_display_widget->Clear();
	d_display_widget->Set_Background_Image(QPixmap::fromImage(d_data.Page_Image()));
	
	for(cntr = 0;cntr < d_data.Item_Count();++cntr)
	{
		data_item = d_data.Item(cntr);
		
		d_display_widget->Add_Item(data_item.id,data_item.value,data_item.text,data_item.border_type);
		d_display_widget->Move_Item(data_item.id,data_item.position);
	}
	
	d_display_widget->setUpdatesEnabled(true);

	return return_state;
}


