/////////////////////////////////////////////////////////////////////
//
//            X   X           X
//           XXX XX         XX
//          XXXXXXXX      XXX
//         XX X XXXXXXXXXXX
//        XXXXX XXXXXXXXX
//       XXXXX XXXXXXXXXX
//            XXX XXX XXX
//           XXX XX   XX
//           X   X     X
//
//    Copyright (C) 2003-2020  Ron Jakl
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
/////////////////////////////////////////////////////////////////////


#include <QApplication>
#include <QCheckBox>
#include <QFrame>
#include <QGridLayout>
#include <QLabel>
#include <QLineEdit>
#include <QSizePolicy>
#include <QSlider>
// #include <QSpacerItem>
#include <QToolButton>
#include <QColorDialog>
#include <QComboBox>
#include <QPixmap>
#include <QResizeEvent>

#include "datatypes.h"

#include "settingswidget.h"

TSettingsWidget::TSettingsWidget(
	const QWidget						*parent,
	const Qt::WindowFlags				flags)
:QWidget(const_cast<QWidget*>(parent),flags)
{
	QFrame								*lower_hseparator_line;
	QFrame								*upper_hseparator_line;
	QGridLayout							*widget_layout;
	QLabel								*item_color_label;
	QLabel								*item_list_format_label;
	QLabel								*item_size_label;
	QLabel								*pdf_orientation_label;
	QLabel								*pdf_resolution_label;
	QSizePolicy							size_policy(QSizePolicy::Minimum,QSizePolicy::Fixed);
	QSpacerItem							*widget_vspacer;

	size_policy.setHorizontalStretch(0);
	size_policy.setVerticalStretch(0);

	this->resize(315,217);

	widget_layout = new QGridLayout(this);

	item_list_format_label = new QLabel(this);
	widget_layout->addWidget(item_list_format_label,0,0,1,1);

	d_item_list_format_edit = new QLineEdit(this);
	d_item_list_format_edit->setAlignment(Qt::AlignCenter);
	widget_layout->addWidget(d_item_list_format_edit,0,1,1,1);

	item_color_label = new QLabel(this);
	widget_layout->addWidget(item_color_label,1,0,1,1);

	d_color_select_button = new QToolButton(this);
	d_color_select_button->setSizePolicy(size_policy);
	widget_layout->addWidget(d_color_select_button,1,1,1,1);

	item_size_label = new QLabel(this);
	widget_layout->addWidget(item_size_label,2,0,1,1);

	d_item_size_slider = new QSlider(this);
	d_item_size_slider->setOrientation(Qt::Horizontal);
	d_item_size_slider->setRange(16,128);
	widget_layout->addWidget(d_item_size_slider,2,1,1,1);
	
	upper_hseparator_line = new QFrame(this);
	upper_hseparator_line->setFrameShape(QFrame::HLine);
	upper_hseparator_line->setFrameShadow(QFrame::Sunken);
	widget_layout->addWidget(upper_hseparator_line,3,0,1,2);

	pdf_orientation_label = new QLabel(this);
	widget_layout->addWidget(pdf_orientation_label,4,0,1,1);
	
	d_pdf_orientation_combo = new QComboBox(this);
	widget_layout->addWidget(d_pdf_orientation_combo,4,1,1,1);
	
	pdf_resolution_label = new QLabel(this);
	widget_layout->addWidget(pdf_resolution_label,5,0,1,1);
	
	d_pdf_resolution_combo = new QComboBox(this);
	widget_layout->addWidget(d_pdf_resolution_combo,5,1,1,1);
	
	lower_hseparator_line = new QFrame(this);
	lower_hseparator_line->setFrameShape(QFrame::HLine);
	lower_hseparator_line->setFrameShadow(QFrame::Sunken);
	widget_layout->addWidget(lower_hseparator_line,6,0,1,2);

	d_track_item_check = new QCheckBox(this);
	widget_layout->addWidget(d_track_item_check,7,0,1,2);

	widget_vspacer = new QSpacerItem(0,0,QSizePolicy::Minimum,QSizePolicy::Expanding);
	widget_layout->addItem(widget_vspacer,8,0,1,2);
	
	// defaults
	d_item_size_slider->setValue(32);
	d_current_color = QColor(TDefaultColor::COLOR_LABEL_RED,TDefaultColor::COLOR_LABEL_GREEN,TDefaultColor::COLOR_LABEL_BLUE);
	d_draw_on_resize = true;
	d_track_item_check->setChecked(true);
	
	this->setWindowTitle(QStringLiteral("Settings"));
	
	item_list_format_label->setText(QStringLiteral("Item Display Format:"));
	item_color_label->setText(QStringLiteral("Item Color:"));
	item_size_label->setText(QStringLiteral("Item Size:"));
	d_track_item_check->setText(QStringLiteral("Track Selected Item"));
	d_item_list_format_edit->setText(QStringLiteral("Item[%1]"));
	pdf_orientation_label->setText(QStringLiteral("PDF Orientation"));
	pdf_resolution_label->setText(QStringLiteral("PDF Resolution"));
	
	d_pdf_orientation_combo->addItem(QStringLiteral("Portrait"),0);
	d_pdf_orientation_combo->addItem(QStringLiteral("Landscape"),1);
	
	d_pdf_resolution_combo->addItem(QStringLiteral("96 DPI"),1);
	d_pdf_resolution_combo->addItem(QStringLiteral("180 DPI"),2);
	d_pdf_resolution_combo->addItem(QStringLiteral("276 DPI"),3);
	d_pdf_resolution_combo->addItem(QStringLiteral("372 DPI"),4);

	connect(d_color_select_button,SIGNAL(clicked(void)),this,SLOT(Select_Color(void)));
	connect(d_item_size_slider,SIGNAL(valueChanged(int)),this,SIGNAL(Label_Size_Changed(int)));
	connect(d_item_list_format_edit,SIGNAL(editingFinished(void)),this,SLOT(Display_Format_Updated(void)));
	connect(d_track_item_check,SIGNAL(toggled(bool)),this,SIGNAL(Track_Item_Option_Changed(bool)));
	
	emit Color_Changed(d_current_color);
}

TSettingsWidget::~TSettingsWidget(void)
{
}

int TSettingsWidget::Label_Size(void) const
{
	return d_item_size_slider->value();
}

QColor TSettingsWidget::Label_Color(void) const
{
	return d_current_color;
}

bool TSettingsWidget::Track_Label(void) const
{
	return d_track_item_check->isChecked();
}

QString TSettingsWidget::Display_Format(void) const
{
	return d_item_list_format_edit->text();
}

int TSettingsWidget::PDF_Orientation(void) const
{
	return d_pdf_orientation_combo->currentData(Qt::UserRole).toInt();
}

int TSettingsWidget::PDF_Resolution(void) const
{
	return d_pdf_resolution_combo->currentData(Qt::UserRole).toInt();
}

void TSettingsWidget::Set_Label_Size(
	const int							size)
{
	d_item_size_slider->setValue(size);
}

void TSettingsWidget::Set_Label_Color(
	const QColor						&color)
{
	d_current_color = color;	
	d_color_select_button->setIcon(this->Create_Icon(d_color_select_button->iconSize(),d_current_color));
}

void TSettingsWidget::Set_Track_Label(
	const bool							state)
{
	d_track_item_check->setChecked(state);
}

void TSettingsWidget::Set_Display_Format(
	const QString						&format)
{	
	d_item_list_format_edit->setText(format);
}
void TSettingsWidget::Set_PDF_Orientation(
	const int							orientation)
{
	if(orientation)
	{
		d_pdf_orientation_combo->setCurrentIndex(1);
	}
	else
	{
		d_pdf_orientation_combo->setCurrentIndex(0);
	}
}

void TSettingsWidget::Set_PDF_Resolution(
	const int							resolution)
{
	switch(resolution)
	{
		case 1:
		default:
			d_pdf_resolution_combo->setCurrentIndex(0);
			break;
			
		case 2:
			d_pdf_resolution_combo->setCurrentIndex(1);
			break;
			
		case 3:
			d_pdf_resolution_combo->setCurrentIndex(2);
			break;
			
		case 4:
			d_pdf_resolution_combo->setCurrentIndex(3);
			break;
	}
}

void TSettingsWidget::Select_Color(void)
{
	QColorDialog						dialog(this);
	
	dialog.setCurrentColor(d_current_color);
	dialog.setOptions(0);
	
	if(dialog.exec() == QDialog::Accepted)
	{
		d_current_color = dialog.currentColor();
		
		d_color_select_button->setIcon(this->Create_Icon(d_color_select_button->iconSize(),d_current_color));
		
		emit Color_Changed(d_current_color);
	}
}

void TSettingsWidget::Display_Format_Updated(void)
{
	QString								text;
	QString								input_text;
	int									index1;
	int									index2;
	
	input_text = d_item_list_format_edit->text();
	text = input_text.simplified();
	
	// make sure it meet the minimum requirements
	
	if(text.length())
	{
		index1 = text.indexOf("%1");
		index2 = text.indexOf("%2");
		
		text.replace('%','_');

		if(!(index1 < 0))
		{
			text[index1] = '%';
			
			if(!(index2 < 0))
			{
				text[index2] = '%';
			}
		}
	}
	else
	{
		text = QStringLiteral("Item [%1] %2");
	}
	
	if(input_text != text)
	{
		d_item_list_format_edit->setText(text);
	}
	
	emit Item_Display_Format_Changed(text);
}

QIcon TSettingsWidget::Create_Icon(
	const QSize							&size,
	const QColor						&color) const
{
	QPixmap								pixmap(size);
	
	pixmap.fill(color);
	
	return QIcon(pixmap);
}

void TSettingsWidget::resizeEvent(
	QResizeEvent						*event)
{
	QSize								icon_size;
	
	QWidget::resizeEvent(event);
	
	if(d_draw_on_resize)
	{
		icon_size = d_color_select_button->size() - QSize(6,6);
		
		if(icon_size.width() < 16)
		{
			icon_size.setWidth(16);
		}
		
		if(icon_size.height() < 16)
		{
			icon_size.setHeight(16);
		}
		
		d_color_select_button->setIconSize(icon_size);
		d_color_select_button->setIcon(this->Create_Icon(icon_size,d_current_color));

		d_draw_on_resize = false;
	}
	
}

